import dspy

from autoflow.llms import ChatModel


def get_dspy_lm_by_chat_model(chat_model: ChatModel) -> dspy.LM:
    return dspy.LM(
        model=chat_model.model,
        max_tokens=chat_model.max_tokens,
        **chat_model.additional_kwargs,
    )



import enum
from typing import Optional, Dict
from datetime import datetime
from uuid import UUID

from llama_index.core.schema import Document as LlamaDocument
from sqlalchemy.dialects.mysql import MEDIUMTEXT
from sqlmodel import (
    Field,
    Column,
    DateTime,
    JSON,
    SQLModel,
)


class DocumentIndexStatus(str, enum.Enum):
    NOT_STARTED = "not_started"
    PENDING = "pending"
    CHUNKING = "chunking"
    COMPLETED = "completed"
    FAILED = "failed"


class VectorSearchIndexStatus(str, enum.Enum):
    NOT_STARTED = "not_started"
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"


class KnowledgeGraphIndexStatus(str, enum.Enum):
    NOT_STARTED = "not_started"
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"


class Document(SQLModel, table=True):
    __tablename__ = "documents"

    id: Optional[int] = Field(default=None, primary_key=True)
    hash: str = Field(max_length=32)
    name: str = Field(max_length=256)
    content: str = Field(sa_column=Column(MEDIUMTEXT))
    mime_type: str = Field(max_length=100)
    meta: Optional[Dict] = Field(default={}, sa_column=Column(JSON))
    data_source_id: UUID = Field(nullable=True)
    knowledge_base_id: UUID = Field(nullable=True)
    created_at: Optional[datetime] = Field(default=None, sa_column=Column(DateTime))
    updated_at: Optional[datetime] = Field(default=None, sa_column=Column(DateTime))

    def __hash__(self) -> int:
        return hash(self.hash)

    def to_llama_document(self) -> LlamaDocument:
        return LlamaDocument(
            id_=str(self.id),
            text=self.content,
            metadata=self.meta,
        )



from typing import Any

from autoflow.datasources import (
    DataSource,
    FileDataSource,
    WebSitemapDataSource,
    WebSinglePageDataSource,
)
from autoflow.schema import DataSourceType


def get_datasource_by_type(type: DataSourceType, config: Any) -> DataSource:
    if type == DataSourceType.FILE:
        return FileDataSource(config)
    elif type == DataSourceType.WEB_SITEMAP:
        return WebSitemapDataSource(config)
    elif type == DataSourceType.WEB_SINGLE_PAGE:
        return WebSinglePageDataSource(config)
    else:
        raise ValueError(f"Unknown datasource type: {type}")



import enum
from datetime import datetime
from typing import Optional, Type
from uuid import UUID

from sqlalchemy import Column, DateTime, JSON, Text, func
from sqlalchemy.orm.decl_api import RegistryType
from sqlmodel import Field, SQLModel
from sqlmodel.main import default_registry, Relationship as SQLRelationship
from tidb_vector.sqlalchemy import VectorType

from autoflow.utils.sql_model import PatchSQLModel
from autoflow.utils.uuid6 import uuid7


class KgIndexStatus(str, enum.Enum):
    NOT_STARTED = "not_started"
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"


def get_chunk_model(
    table_name: str,
    vector_dimension: int,
    document_table_name: str,
    document_db_model: Type[SQLModel],
    registry: RegistryType = default_registry,
) -> Type[SQLModel]:
    class Chunk(PatchSQLModel, table=True, registry=registry):
        __tablename__ = table_name
        __table_args__ = {"extend_existing": True}

        id: UUID = Field(primary_key=True, default_factory=uuid7)
        hash: str = Field(max_length=64)
        text: str = Field(sa_column=Column(Text))
        text_vec: list[float] = Field(sa_column=Column(VectorType(vector_dimension)))
        meta: dict = Field(default={}, sa_column=Column(JSON))
        document_id: int = Field(foreign_key=f"{document_table_name}.id", nullable=True)
        document: document_db_model = SQLRelationship()

        kg_index_status: KgIndexStatus = KgIndexStatus.NOT_STARTED
        kg_index_result: str = Field(sa_column=Column(Text, nullable=True))

        created_at: Optional[datetime] = Field(
            default=None,
            sa_column=Column(DateTime(timezone=True), server_default=func.now()),
        )
        updated_at: Optional[datetime] = Field(
            default=None,
            sa_column=Column(
                DateTime(timezone=True), server_default=func.now(), onupdate=func.now()
            ),
        )

    return Chunk



from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional, Generic, TypeVar, Type

from pydantic import BaseModel
from autoflow.models.document import Document
from autoflow.storage.schema import QueryBundle

D = TypeVar("D", bound=Type[Document])
C = TypeVar("C")


class DocumentSearchMethod(str, Enum):
    VECTOR_SEARCH = "vector_search"
    FULLTEXT_SEARCH = "fulltext_search"


class DocumentSearchQuery(QueryBundle):
    # document_ids: Optional[List[int]] = None
    # chunk_ids: Optional[List[int]] = None
    # metadata_filters: Optional[MetadataFilters] = None
    search_method: List[DocumentSearchMethod] = [DocumentSearchMethod.VECTOR_SEARCH]
    top_k: Optional[int] = None

    # Vector Search
    similarity_threshold: Optional[float] = None
    # similarity_weight: Optional[float] = None
    similarity_nprobe: Optional[int] = None
    similarity_top_k: Optional[int] = 5

    # Full Text Search
    # TODO: Support Full Text Search.

    # Reranking
    enable_reranker: bool = False
    reranker_model_name: Optional[str] = None


class ChunkWithScore(BaseModel, Generic[C]):
    chunk: C
    score: float


class DocumentSearchResult(BaseModel, Generic[D, C]):
    chunks: List[ChunkWithScore[C]]
    documents: List[Document]


class DocumentStore(ABC, Generic[D, C]):
    @abstractmethod
    def add(self, document: List[D]) -> List[D]:
        raise NotImplementedError()

    @abstractmethod
    def delete(self, document_id: int) -> None:
        raise NotImplementedError()

    @abstractmethod
    def search(self, query: DocumentSearchQuery) -> DocumentSearchResult[D, C]:
        raise NotImplementedError()

    @abstractmethod
    def get(self, document_id: int) -> D:
        raise NotImplementedError()

    @abstractmethod
    def add_doc_chunks(self, chunks: List[C]) -> List[C]:
        raise NotImplementedError()



import logging
from typing import Optional, List, Any, Dict

import sqlalchemy
from pydantic import BaseModel
from sqlalchemy import Engine, update, text
from sqlalchemy.orm import Session, DeclarativeMeta
from sqlmodel.main import SQLModelMetaclass
from tidb_vector.sqlalchemy import VectorAdaptor

from autoflow.storage.tidb.base import Base
from autoflow.storage.tidb.constants import VectorDataType, TableModel, DistanceMetric
from autoflow.storage.tidb.query import QueryType, TiDBVectorQuery
from autoflow.storage.tidb.utils import (
    build_filter_clauses,
    check_vector_column,
    filter_vector_columns,
)

logger = logging.getLogger(__name__)


class Table:
    def __init__(
        self,
        *,
        db_engine: Engine,
        schema: Optional[TableModel] = None,
        vector_column: Optional[str] = None,
        distance_metric: Optional[DistanceMetric] = DistanceMetric.COSINE,
        checkfirst: bool = True,
    ):
        self._db_engine = db_engine

        # Init table model.
        if type(schema) is SQLModelMetaclass:
            self._table_model = schema
        elif type(schema) is DeclarativeMeta:
            self._table_model = schema
        else:
            raise TypeError(f"Invalid schema type: {type(schema)}")
        self._columns = self._table_model.__table__.columns

        # Field for auto embedding.
        self._vector_field_configs = {}
        for name, field in schema.__pydantic_fields__.items():
            # FIXME: using field custom attributes instead of it.
            if "embed_fn" in field._attributes_set:
                embed_fn = field._attributes_set["embed_fn"]
                source_field_name = field._attributes_set["source_field"]
                self._vector_field_configs[name] = {
                    "embed_fn": embed_fn,
                    "vector_field": field,
                    "source_field_name": source_field_name,
                }

        # Create table.
        Base.metadata.create_all(self._db_engine, tables=[self._table_model.__table__], checkfirst=checkfirst)

        # Create index.
        self._vector_columns = filter_vector_columns(self._columns)
        vector_adaptor = VectorAdaptor(self._db_engine)
        for col in self._vector_columns:
            if vector_adaptor.has_vector_index(col):
                continue
            vector_adaptor.create_vector_index(col, distance_metric)

        # Determine vector column for search.
        if vector_column is not None:
            self._vector_column = check_vector_column(self._columns, vector_column)
        else:
            if len(self._vector_columns) == 1:
                self._vector_column = self._vector_columns[0]
            else:
                self._vector_column = None

    @property
    def table_model(self) -> TableModel:
        return self._table_model

    @property
    def table_name(self) -> str:
        return self._table_model.__tablename__

    @property
    def db_engine(self) -> Engine:
        return self._db_engine

    @property
    def vector_column(self):
        return self._vector_column

    @property
    def vector_columns(self):
        return self._vector_columns

    @property
    def vector_field_configs(self):
        return self._vector_field_configs

    def get(self, id: int):
        with Session(self._db_engine) as session:
            return session.get(self._table_model, id)

    def insert(self, data: BaseModel):
        # Auto embedding.
        for field_name, config in self._vector_field_configs.items():
            if getattr(data, field_name) is not None:
                # Vector embeddings is provided.
                continue

            if not hasattr(data, config["source_field_name"]):
                continue

            embedding_source = getattr(data, config["source_field_name"])
            vector_embedding = config["embed_fn"].get_source_embedding(embedding_source)
            setattr(data, field_name, vector_embedding)

        with Session(self._db_engine) as session:
            session.add(data)
            session.commit()
            session.refresh(data)
            return data

    def bulk_insert(self, data: List[object]) -> List[object]:
        # Auto embedding.
        for field_name, config in self._vector_field_configs.items():
            items_need_embedding = []
            sources_to_embedding = []
            for item in data:
                if getattr(item, field_name) is not None:
                    continue
                if not hasattr(item, config["source_field_name"]):
                    continue
                items_need_embedding.append(item)
                embedding_source = getattr(item, config["source_field_name"])
                sources_to_embedding.append(embedding_source)

            vector_embeddings = config["embed_fn"].get_source_embedding_batch(
                sources_to_embedding
            )
            for item, embedding in zip(items_need_embedding, vector_embeddings):
                setattr(item, field_name, embedding)

        with Session(self._db_engine) as session:
            session.add_all(data)
            session.commit()
            for item in data:
                session.refresh(item)
            return data

    def update(self, values: dict, filters: Optional[Dict[str, Any]] = None) -> object:
        filter_clauses = build_filter_clauses(filters, self._columns, self._table_model)
        with Session(self._db_engine) as session:
            stmt = update(self._table_model).filter(*filter_clauses).values(values)
            session.execute(stmt)
            session.commit()

    def delete(self, filters: Optional[Dict[str, Any]] = None):
        """
        Delete data from the TiDB table.

        params:
            filters: (Optional[Dict[str, Any]]): The filters to apply to the delete operation.
        """
        filter_clauses = build_filter_clauses(filters, self._columns, self._table_model)
        with Session(self._db_engine) as session:
            stmt = sqlalchemy.delete(self._table_model).filter(*filter_clauses)
            session.execute(stmt)
            session.commit()

    def truncate(self):
        with Session(self._db_engine) as session:
            stmt = text(f"TRUNCATE TABLE {self.table_name}")
            session.execute(stmt)

    def rows(self):
        with Session(self._db_engine) as session:
            stmt = text(f"SELECT COUNT(*) FROM {self.table_name}")
            res = session.execute(stmt)
            return res.scalar()

    def query(self, filters: Optional[Dict[str, Any]] = None):
        with Session(self._db_engine) as session:
            query = session.query(self._table_model)
            if filters:
                filter_clauses = build_filter_clauses(
                    filters, self._columns, self._table_model
                )
                query = query.filter(*filter_clauses)
            return query.all()

    def search(
        self,
        query: VectorDataType | str,
        query_type: QueryType = QueryType.VECTOR_SEARCH,
    ):
        if query_type == QueryType.VECTOR_SEARCH:
            return TiDBVectorQuery(
                table=self,
                query=query,
            )
        else:
            raise ValueError(f"Unsupported query type: {query_type}")



import enum


class SupportedMimeTypes(str, enum.Enum):
    PLAIN_TXT = "text/plain"
    MARKDOWN = "text/markdown"
    PDF = "application/pdf"
    DOCX = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    PPTX = "application/vnd.openxmlformats-officedocument.presentationml.presentation"
    XLSX = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    CSV = "text/csv"



from .base import DocumentStore, DocumentSearchQuery, DocumentSearchResult
from .tidb.tidb_doc_store import TiDBDocumentStore

__all__ = [
    "DocumentStore",
    "TiDBDocumentStore",
    "DocumentSearchQuery",
    "DocumentSearchResult",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from uuid import UUID
from pathlib import Path

import streamlit as st  # type: ignore
from sqlalchemy import create_engine
from autoflow import Autoflow
from autoflow.schema import IndexMethod
from autoflow.llms.chat_models import ChatModel
from autoflow.llms.embeddings import EmbeddingModel
from llama_index.core.llms import ChatMessage

st.set_page_config(
    page_title="Demo of Autoflow and Streamlit", page_icon="📖", layout="wide"
)
st.header("📖 Knowledge base app built with Autoflow and Streamlit")

with st.sidebar:
    st.markdown(
        "## How to use\n"
        "1. Enter your [OpenAI API key](https://platform.openai.com/account/api-keys) below 🔑\n"  # noqa: E501
        "2. Enter your [TiDB Cloud](https://tidbcloud.com) database connection URL below 🔗\n"
        "3. Upload a pdf, docx, or txt file 📄\n"
        "4. Ask a question about the document 💬\n"
    )
    openai_api_key_input = st.text_input(
        "OpenAI API Key",
        type="password",
        placeholder="Paste your OpenAI API key here (sk-...)",
        help="You can get your API key from https://platform.openai.com/account/api-keys.",  # noqa: E501
        value=os.environ.get("OPENAI_API_KEY", None)
        or st.session_state.get("OPENAI_API_KEY", ""),
    )
    database_url_input = st.text_input(
        "Database URL",
        type="password",
        placeholder="e.g. mysql+pymysql://root@localhost:4000/test",
        autocomplete="off",
        help="You can get your database URL from https://tidbcloud.com",
        value=os.environ.get("DATABASE_URL", None)
        or "mysql+pymysql://root@localhost:4000/test"
        or st.session_state.get("DATABASE_URL", ""),
    )
    st.session_state["OPENAI_API_KEY"] = openai_api_key_input
    st.session_state["DATABASE_URL"] = database_url_input

openai_api_key = st.session_state.get("OPENAI_API_KEY")
database_url = st.session_state.get("DATABASE_URL")

if not openai_api_key or not database_url:
    st.error("Please enter your OpenAI API key and TiDB Cloud connection string.")
    st.stop()

af = Autoflow(create_engine(database_url))
chat_model = ChatModel("gpt-4o-mini", api_key=openai_api_key)
embedding_model = EmbeddingModel(
    model_name="text-embedding-3-small",
    dimensions=1536,
    api_key=openai_api_key,
)
kb = af.create_knowledge_base(
    id=UUID(
        "655b6cf3-8b30-4839-ba8b-5ed3c502f30e"
    ),  # For not creating a new KB every time
    name="New KB",
    description="This is a knowledge base for testing",
    index_methods=[IndexMethod.VECTOR_SEARCH, IndexMethod.KNOWLEDGE_GRAPH],
    chat_model=chat_model,
    embedding_model=embedding_model,
)

with st.form(key="file_upload_form"):
    uploaded_file = st.file_uploader(
        "Upload a .pdf, .docx, .md or .txt file",
        type=["pdf", "docx", "txt", "md"],
        help="Scanned documents are not supported yet!",
    )
    upload = st.form_submit_button("Upload")
    if upload:
        if not uploaded_file:
            st.error("Please upload a valid file.")
            st.stop()
        file_path = f"/tmp/{uploaded_file.name}"
        with st.spinner(
            "Indexing document... This may take a while ⏳(import time; time.sleep(3))"
        ):
            with open(file_path, "wb") as f:
                f.write(uploaded_file.getvalue())
            kb.import_documents_from_files(
                files=[
                    Path(file_path),
                ]
            )
            import time

            time.sleep(3)

for line in ["generated", "past", "corpus"]:
    if line not in st.session_state:
        st.session_state[line] = []

for o in ["kg"]:
    if o not in st.session_state:
        st.session_state[o] = None


def on_submit():
    user_input = st.session_state.user_input
    if user_input:
        result = kb.search_documents(query=user_input, similarity_top_k=3)
        st.session_state["corpus"] = result.chunks
        kg = kb.search_knowledge_graph(query=user_input)
        st.session_state["kg"] = kg
        messages = [
            ChatMessage(
                role="system",
                content="Here are some relevant documents about your query:\n\n"
                + "\n".join(c.chunk.text for c in result.chunks),
            ),
            ChatMessage(
                role="user",
                content=user_input + "\n(in markdown, removed unused breaklines)",
            ),
        ]
        resp = chat_model.chat(messages)
        st.session_state["past"].append(user_input)
        st.session_state.generated.append(str(resp.message))


chat_section, corpus_section = st.columns(2)
with chat_section:
    st.markdown("##### Chats")
    chat_placeholder = st.empty()
    with chat_placeholder.container():
        for i in range(len(st.session_state["generated"])):
            with st.chat_message("user"):
                st.write(st.session_state["past"][i])
            with st.chat_message("assistant"):
                st.write(st.session_state["generated"][i])

    with st.container():
        st.chat_input(
            "Input your question about this document here.",
            key="user_input",
            on_submit=on_submit,
        )

with corpus_section:
    st.markdown("##### Vector Search Results")
    corpus_placeholder = st.empty()
    with corpus_placeholder.container():
        [c.chunk for c in st.session_state["corpus"]] if st.session_state[
            "corpus"
        ] else "Please input a query left."

    st.markdown("##### Knowledge Graph Search Results")
    kg_placeholder = st.empty()
    with kg_placeholder.container():
        kg = st.session_state["kg"]
        [
            r.rag_description for r in kg.relationships
        ] if kg else "Please input a query left."



import json
import logging
from collections import defaultdict
from contextlib import contextmanager

import numpy as np
import tidb_vector
import sqlalchemy

from typing import List, Optional, Tuple, Dict, Type, Sequence, Any, Mapping, Collection

from deepdiff import DeepDiff
from sqlalchemy.orm import defer, joinedload
from sqlalchemy.orm.attributes import flag_modified
from sqlmodel import Session, asc, func, select, text, or_, SQLModel
from tidb_vector.sqlalchemy import VectorAdaptor
from sqlalchemy import desc, Engine

from autoflow.indices.knowledge_graph.schema import (
    AIKnowledgeGraph,
    ProcessedAIKnowledgeGraph,
    AIRelationshipWithEntityDesc,
    AIEntity,
)
from autoflow.llms import EmbeddingModel
from autoflow.models.entity import EntityType
from autoflow.storage.graph_store import KnowledgeGraphStore
from autoflow.storage.graph_store.base import (
    EntityFilters,
    EntityCreate,
    EntityUpdate,
    EntityDegree,
    RelationshipUpdate,
    RelationshipFilters,
    RetrievedEntity,
    RetrievedRelationship,
    E,
    R,
    C,
)
from autoflow.storage.schema import QueryBundle

logger = logging.getLogger(__name__)


def cosine_distance(v1, v2):
    return 1 - np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))


def _process_ai_knowledge_graph(
    knowledge_graph: AIKnowledgeGraph, chunk_metadata: Mapping[str, Any]
) -> ProcessedAIKnowledgeGraph:
    entities = knowledge_graph.entities
    relationships = []
    mapped_entities = {entity.name: entity for entity in entities}

    for relationship in knowledge_graph.relationships:
        if relationship.source_entity not in mapped_entities:
            new_source_entity = AIEntity(
                name=relationship.source_entity,
                description=(
                    f"Derived from from relationship: "
                    f"{relationship.source_entity} -> {relationship.relationship_desc} -> {relationship.target_entity}"
                ),
                metadata={"status": "need-revised"},
            )
            entities.append(new_source_entity)
            mapped_entities[relationship.source_entity] = new_source_entity
            source_entity_description = new_source_entity.description
        else:
            source_entity_description = mapped_entities[
                relationship.source_entity
            ].description

        if relationship.target_entity not in mapped_entities:
            new_target_entity = AIEntity(
                name=relationship.target_entity,
                description=(
                    f"Derived from from relationship: "
                    f"{relationship.source_entity} -> {relationship.relationship_desc} -> {relationship.target_entity}"
                ),
                metadata={"status": "need-revised"},
            )
            entities.append(new_target_entity)
            mapped_entities[relationship.target_entity] = new_target_entity
            target_entity_description = new_target_entity.description
        else:
            target_entity_description = mapped_entities[
                relationship.target_entity
            ].description

        relationships.append(
            AIRelationshipWithEntityDesc(
                source_entity=relationship.source_entity,
                source_entity_description=source_entity_description,
                target_entity=relationship.target_entity,
                target_entity_description=target_entity_description,
                relationship_desc=relationship.relationship_desc,
                meta={**chunk_metadata},
            )
        )

    return ProcessedAIKnowledgeGraph(
        entities=entities,
        relationships=relationships,
    )


class TiDBKnowledgeGraphStore(KnowledgeGraphStore[E, R, C]):
    def __init__(
        self,
        db_engine: Engine,
        embedding_model: EmbeddingModel,
        entity_db_model: Type[E],
        relationship_db_model: Type[R],
        entity_similarity_threshold: Optional[float] = 0.9,
    ):
        super().__init__(embedding_model)
        self._db_engine = db_engine
        self._embedding_model = embedding_model
        self._entity_db_model = entity_db_model
        self._relationship_db_model = relationship_db_model
        self._entity_distance_threshold = 1 - entity_similarity_threshold

    # Schema Operations

    # TODO: move to low-level storage API.
    def ensure_table_schema(self) -> None:
        inspector = sqlalchemy.inspect(self._db_engine)
        existed_table_names = inspector.get_table_names()

        entity_model = self._entity_db_model
        entities_table_name = entity_model.__tablename__
        if entities_table_name not in existed_table_names:
            entity_model.metadata.create_all(
                self._db_engine, tables=[entity_model.__table__]
            )

            # Add HNSW index to accelerate ann queries.
            VectorAdaptor(self._db_engine).create_vector_index(
                entity_model.description_vec, tidb_vector.DistanceMetric.COSINE
            )
            VectorAdaptor(self._db_engine).create_vector_index(
                entity_model.meta_vec, tidb_vector.DistanceMetric.COSINE
            )

            logger.info(
                f"Entities table <{entities_table_name}> has been created successfully."
            )
        else:
            logger.info(
                f"Entities table <{entities_table_name}> is already exists, not action to do."
            )

        relationship_model = self._relationship_db_model
        relationships_table_name = relationship_model.__tablename__
        if relationships_table_name not in existed_table_names:
            relationship_model.metadata.create_all(
                self._db_engine, tables=[relationship_model.__table__]
            )

            # Add HNSW index to accelerate ann queries.
            VectorAdaptor(self._db_engine).create_vector_index(
                relationship_model.description_vec,
                tidb_vector.DistanceMetric.COSINE,
            )

            logger.info(
                f"Relationships table <{relationships_table_name}> has been created successfully."
            )
        else:
            logger.info(
                f"Relationships table <{relationships_table_name}> is already exists, not action to do."
            )

    def drop_table_schema(self) -> None:
        inspector = sqlalchemy.inspect(self._db_engine)
        existed_table_names = inspector.get_table_names()
        relationships_table_name = self._relationship_db_model.__tablename__
        entities_table_name = self._entity_db_model.__tablename__

        if relationships_table_name in existed_table_names:
            self._relationship_db_model.metadata.drop_all(
                self._db_engine, tables=[self._relationship_db_model.__table__]
            )
            logger.info(
                f"Relationships table <{relationships_table_name}> has been dropped successfully."
            )
        else:
            logger.info(
                f"Relationships table <{relationships_table_name}> is not existed, not action to do."
            )

        if entities_table_name in existed_table_names:
            self._entity_db_model.metadata.drop_all(
                self._db_engine, tables=[self._entity_db_model.__table__]
            )
            logger.info(
                f"Entities table <{entities_table_name}> has been dropped successfully."
            )
        else:
            logger.info(
                f"Entities table <{entities_table_name}> is not existed, not action to do."
            )

    @contextmanager
    def _session_scope(self, session: Optional[Session] = None, commit: bool = False):
        """Provide a transactional scope around a series of operations."""
        should_close = session is None
        session = session or Session(self._db_engine)

        try:
            yield session
            if commit:
                session.commit()
            else:
                session.flush()
        except Exception as e:
            session.rollback()
            raise e
        finally:
            if should_close:
                session.close()

    def _get_entity_description_embedding(
        self, name: str, description: str
    ) -> List[float]:
        # TODO: Make it configurable.
        embedding_str = f"{name}: {description}"
        return self._embedding_model.get_text_embedding(embedding_str)

    def _get_entity_metadata_embedding(self, metadata: Dict[str, Any]) -> List[float]:
        embedding_str = json.dumps(metadata, ensure_ascii=False)
        return self._embedding_model.get_text_embedding(embedding_str)

    def _get_relationship_description_embedding(
        self,
        source_entity_name: str,
        source_entity_description,
        target_entity_name: str,
        target_entity_description: str,
        relationship_desc: str,
    ) -> List[float]:
        # TODO: Make it configurable.
        embedding_str = (
            f"{source_entity_name}({source_entity_description}) -> "
            f"{relationship_desc} -> {target_entity_name}({target_entity_description}) "
        )
        return self._embedding_model.get_text_embedding(embedding_str)

    def save_knowledge_graph(
        self, knowledge_graph: AIKnowledgeGraph, chunk: SQLModel
    ) -> None:
        if (
            len(knowledge_graph.entities) == 0
            or len(knowledge_graph.relationships) == 0
        ):
            logger.info(
                "Entities or relationships are empty, skip saving to the database"
            )
            return

        chunk_id = chunk.id
        exists_relationships = self.list_relationships(
            RelationshipFilters(chunk_ids=[chunk_id])
        )
        if len(exists_relationships) > 0:
            logger.info(f"{chunk_id} already exists in the relationship table, skip.")
            return

        knowledge_graph = _process_ai_knowledge_graph(knowledge_graph, chunk.meta)

        with self._session_scope() as session:
            entities_name_map = defaultdict(list)
            for entity in knowledge_graph.entities:
                entities_name_map[entity.name].append(
                    self.find_or_create_entity(
                        EntityCreate(
                            entity_type=EntityType.original,
                            name=entity.name,
                            description=entity.description,
                            meta=entity.metadata,
                        ),
                        commit=False,
                        db_session=session,
                    )
                )

            def _find_or_create_entity_for_relation(
                name: str, description: str
            ) -> SQLModel:
                _embedding = self._get_entity_description_embedding(name, description)

                # Check entities_name_map first, if not found, then check the database.
                entities_with_same_name = entities_name_map.get(name, [])
                if len(entities_with_same_name) > 0:
                    the_closest_entity = None
                    the_closest_distance = float("inf")
                    for en in entities_with_same_name:
                        distance = cosine_distance(en.description_vec, _embedding)
                        if distance < the_closest_distance:
                            the_closest_distance = distance
                            the_closest_entity = en
                    if the_closest_distance < self._entity_distance_threshold:
                        return the_closest_entity

                return self.find_or_create_entity(
                    EntityCreate(
                        entity_type=EntityType.original,
                        name=name,
                        description=description,
                        meta=entity.metadata,
                    ),
                    commit=False,
                    db_session=session,
                )

            for r in knowledge_graph.relationships:
                logger.info(
                    "Save entities for relationship: %s -> %s -> %s",
                    r.source_entity,
                    r.relationship_desc,
                    r.target_entity,
                )
                source_entity = _find_or_create_entity_for_relation(
                    r.source_entity, r.source_entity_description
                )
                target_entity = _find_or_create_entity_for_relation(
                    r.target_entity, r.target_entity_description
                )

                self.create_relationship(
                    source_entity=source_entity,
                    target_entity=target_entity,
                    description=r.relationship_desc,
                    commit=False,
                    db_session=session,
                )
            session.commit()

    # Entity Basic Operations

    def list_entities(
        self,
        filters: Optional[EntityFilters] = EntityFilters(),
        db_session: Session = None,
    ) -> Sequence[E]:
        with self._session_scope(db_session) as session:
            stmt = self._build_entities_query(filters)
            return session.exec(stmt).all()

    def _build_entities_query(self, filters: EntityFilters):
        stmt = select(self._entity_db_model)
        if filters.entity_type:
            stmt = stmt.where(self._entity_db_model.entity_type == filters.entity_type)
        if filters.search:
            stmt = stmt.where(
                or_(
                    self._entity_db_model.name.like(f"%{filters.search}%"),
                    self._entity_db_model.description.like(f"%{filters.search}%"),
                )
            )
        return stmt

    def get_entity_by_id(self, entity_id: int) -> Type[E]:
        with self._session_scope() as session:
            return session.get(self._entity_db_model, entity_id)

    def must_get_entity_by_id(self, entity_id: int) -> Type[E]:
        entity = self.get_entity_by_id(entity_id)
        if entity is None:
            raise ValueError(f"Entity <{entity_id}> does not exist")
        return entity

    def create_entity(
        self, create: EntityCreate, commit: bool = True, db_session: Session = None
    ) -> E:
        desc_vec = self._get_entity_description_embedding(
            create.name, create.description
        )
        meta_vec = self._get_entity_metadata_embedding(create.meta)
        entity = self._entity_db_model(
            name=create.name,
            entity_type=EntityType.original,
            description=create.description,
            description_vec=desc_vec,
            meta=create.meta,
            meta_vec=meta_vec,
        )

        with self._session_scope(db_session, commit) as session:
            session.add(entity)
        return entity

    def find_or_create_entity(
        self, create: EntityCreate, commit: bool = True, db_session: Session = None
    ) -> E:
        most_similar_entity = self._get_the_most_similar_entity(
            create, db_session=db_session
        )

        if most_similar_entity is not None:
            return most_similar_entity

        return self.create_entity(create, commit=commit, db_session=db_session)

    def update_entity(
        self,
        entity: Type[E],
        update: EntityUpdate,
        commit: bool = True,
        db_session: Session = None,
    ) -> Type[E]:
        for key, value in update.model_dump().items():
            if value is None:
                continue
            setattr(entity, key, value)
            flag_modified(entity, key)

        entity.description_vec = self._get_entity_description_embedding(
            entity.name, entity.description
        )
        if update.meta is not None:
            entity.meta_vec = self._get_entity_metadata_embedding(entity.meta)

        with self._session_scope(db_session, commit) as session:
            session.add(entity)
            # Update linked relationships.
            connected_relationships = self.list_entity_relationships(entity.id)
            for relationship in connected_relationships:
                self.update_relationship(relationship, RelationshipUpdate(), commit)
            session.refresh(entity)
            return entity

    def delete_entity(
        self, entity: Type[E], commit: bool = True, db_session: Session = None
    ) -> None:
        with self._session_scope(db_session, commit) as session:
            # Delete linked relationships.
            linked_relationships = self.list_entity_relationships(entity.id)
            for relationship in linked_relationships:
                session.delete(relationship)

            session.delete(entity)

    def list_entity_relationships(self, entity_id: int) -> List[R]:
        stmt = (
            select(self._relationship_db_model)
            .options(
                defer(self._relationship_db_model.description_vec),
                joinedload(self._relationship_db_model.source_entity)
                .defer(self._entity_db_model.description_vec)
                .defer(self._entity_db_model.meta_vec),
                joinedload(self._relationship_db_model.target_entity)
                .defer(self._entity_db_model.description_vec)
                .defer(self._entity_db_model.meta_vec),
            )
            .where(
                (self._relationship_db_model.source_entity_id == entity_id)
                | (self._relationship_db_model.target_entity_id == entity_id)
            )
        )
        with self._session_scope() as session:
            return session.exec(stmt).all()

    def calc_entity_out_degree(self, entity_id: int) -> Optional[int]:
        stmt = select(func.count(self._relationship_db_model.id)).where(
            self._relationship_db_model.source_entity_id == entity_id
        )
        with self._session_scope() as session:
            return session.exec(stmt).one()

    def calc_entity_in_degree(self, entity_id: int) -> Optional[int]:
        stmt = select(func.count(self._relationship_db_model.id)).where(
            self._relationship_db_model.target_entity_id == entity_id
        )
        with self._session_scope() as session:
            return session.exec(stmt).one()

    def calc_entity_degree(self, entity_id: int) -> Optional[int]:
        stmt = select(func.count(self._relationship_db_model.id)).where(
            or_(
                self._relationship_db_model.target_entity_id == entity_id,
                self._relationship_db_model.source_entity_id == entity_id,
            )
        )
        with self._session_scope() as session:
            return session.exec(stmt).one()

    def bulk_calc_entities_degrees(
        self, entity_ids: Collection[int]
    ) -> Mapping[int, EntityDegree]:
        stmt = (
            select(
                self._entity_db_model.id,
                func.count(self._relationship_db_model.id)
                .filter(
                    self._relationship_db_model.source_entity_id
                    == self._entity_db_model.id
                )
                .label("out_degree"),
                func.count(self._relationship_db_model.id)
                .filter(
                    self._relationship_db_model.target_entity_id
                    == self._entity_db_model.id
                )
                .label("in_degree"),
            )
            .where(self._entity_db_model.id.in_(entity_ids))
            .outerjoin(self._relationship_db_model)
            .group_by(self._entity_db_model.id)
        )

        with self._session_scope() as session:
            results = session.exec(stmt).all()
            return {
                item.id: EntityDegree(
                    in_degree=item.in_degree,
                    out_degree=item.out_degree,
                    degrees=item.in_degree + item.out_degree,
                )
                for item in results
            }

    # Entities Retrieve Operations

    def retrieve_entities(
        self,
        query: str,
        entity_type: EntityType = EntityType.original,
        top_k: int = 10,
        nprobe: Optional[int] = None,
        similarity_threshold: Optional[float] = None,
    ) -> List[RetrievedEntity]:
        entities = self.search_similar_entities(
            query=QueryBundle(query_str=query),
            top_k=top_k,
            nprobe=nprobe,
            entity_type=entity_type,
            similarity_threshold=similarity_threshold,
        )
        return [
            RetrievedEntity(
                id=entity.id,
                entity_type=entity.entity_type,
                name=entity.name,
                description=entity.description,
                meta=entity.meta,
                similarity_score=similarity_score,
            )
            for entity, similarity_score in entities
        ]

    def search_similar_entities(
        self,
        query: QueryBundle,
        top_k: int = 10,
        nprobe: Optional[int] = None,
        entity_type: EntityType = EntityType.original,
        similarity_threshold: Optional[float] = None,
        # TODO: Metadata filter
        # TODO: include_metadata, include_metadata_keys, include_embeddings parameters
        db_session: Session = None,
    ) -> List[Tuple[E, float]]:
        if query.query_embedding is None:
            query.query_embedding = self._embedding_model.get_query_embedding(
                query.query_str
            )

        entity_model = self._entity_db_model
        with self._session_scope(db_session) as session:
            if entity_type == EntityType.synopsis:
                return self._search_similar_synopsis_entities(
                    session,
                    entity_model,
                    query.query_embedding,
                    top_k,
                    similarity_threshold,
                )
            else:
                return self._search_similar_original_entities(
                    session,
                    entity_model,
                    query.query_embedding,
                    top_k,
                    nprobe,
                    similarity_threshold,
                )

    def _search_similar_original_entities(
        self,
        db_session: Session,
        entity_model: E,
        query_embedding: List[float],
        top_k: int,
        nprobe: Optional[int] = None,
        similarity_threshold: Optional[float] = None,
    ) -> List[Tuple[E, float]]:
        """
        For original entities, it leverages TiFlash's ANN search to efficiently retrieve the most similar entities
        from a large-scale vector space.

        To optimize retrieval performance on ANN Index, there employ a two-phase retrieval strategy:
        1. Fetch more (nprobe) results from the ANN Index as candidates.
        2. Sort the candidates by distance and get the top-k results.
        """
        nprobe = nprobe or top_k * 10
        subquery = (
            select(
                entity_model.id,
                entity_model.description_vec.cosine_distance(query_embedding).label(
                    "distance"
                ),
            )
            .order_by(asc("distance"))
            .limit(nprobe)
            .subquery("candidates")
        )
        query = (
            select(entity_model, (1 - subquery.c.distance).label("similarity_score"))
            .where(entity_model.id == subquery.c.id)
            .where(entity_model.entity_type == EntityType.original)
        )

        if similarity_threshold is not None:
            distance_threshold = 1 - similarity_threshold
            query = query.where(subquery.c.distance <= distance_threshold)

        query = query.order_by(desc("similarity_score")).limit(top_k)
        return db_session.exec(query).all()

    def _search_similar_synopsis_entities(
        self,
        db_session: Session,
        entity_model: E,
        query_embedding: List[float],
        top_k: int,
        similarity_threshold: Optional[float] = None,
    ) -> List[Tuple[E, float]]:
        """
        For synopsis entities, it leverages TiKV to fetch the synopsis entity quickly by filtering by entity_type,
        because the number of synopsis entities is very small, it is commonly faster than using TiFlash to perform
        ANN search.
        """
        hint = text(f"/*+ read_from_storage(tikv[{entity_model.__tablename__}]) */")
        subquery = (
            select(
                entity_model,
                entity_model.description_vec.cosine_distance(query_embedding).label(
                    "distance"
                ),
            )
            .prefix_with(hint)
            .where(entity_model.entity_type == EntityType.synopsis)
            .order_by(asc("distance"))
            .limit(top_k)
            .subquery("candidates")
        )
        query = select(
            entity_model, (1 - subquery.c.distance).label("similarity_score")
        )

        if similarity_threshold is not None:
            distance_threshold = 1 - similarity_threshold
            query = query.where(subquery.c.distance <= distance_threshold)

        query = query.order_by(desc("similarity_score")).limit(top_k)
        return db_session.exec(query).all()

    def _get_the_most_similar_entity(
        self,
        create: EntityCreate,
        similarity_threshold: float = 0,
        db_session: Optional[Session] = None,
    ) -> Optional[E]:
        query = f"{create.name}: {create.description}"
        similar_entities = self.search_similar_entities(
            QueryBundle(query_str=query), top_k=1, nprobe=10, db_session=db_session
        )

        if len(similar_entities) == 0:
            return None

        most_similar_entity, similarity_score = similar_entities[0]

        # For entity with same name and description.
        if (
            most_similar_entity.name == create.name
            and most_similar_entity.description == create.description
            and len(DeepDiff(most_similar_entity.meta, create.meta)) == 0
        ):
            return most_similar_entity

        # For the most similar entity.
        if similarity_score < similarity_threshold:
            return most_similar_entity

        return None

    # Relationship Basic Operations

    def get_relationship_by_id(
        self, relationship_id: int, db_session: Session = None
    ) -> R:
        stmt = select(self._relationship_db_model).where(
            self._relationship_db_model.id == relationship_id
        )
        with self._session_scope(db_session) as session:
            return session.exec(stmt).first()

    def list_relationships(
        self, filters: RelationshipFilters, db_session: Session = None
    ) -> Sequence[R]:
        stmt = self._build_relationships_query(filters)

        with self._session_scope(db_session) as session:
            return session.exec(stmt).all()

    def _build_relationships_query(self, filters: RelationshipFilters):
        stmt = select(self._relationship_db_model)
        if filters.target_entity_id:
            stmt = stmt.where(
                self._relationship_db_model.target_entity_id == filters.target_entity_id
            )
        if filters.target_entity_id:
            stmt = stmt.where(
                self._relationship_db_model.source_target_id == filters.source_target_id
            )
        if filters.relationship_ids:
            stmt = stmt.where(
                self._relationship_db_model.id.in_(filters.relationship_ids)
            )
        if filters.search:
            stmt = stmt.where(
                or_(
                    self._relationship_db_model.name.like(f"%{filters.search}%"),
                    self._relationship_db_model.description.like(f"%{filters.search}%"),
                )
            )
        if filters.chunk_ids:
            stmt = stmt.where(
                self._relationship_db_model.chunk_id.in_(filters.chunk_ids)
            )
        return stmt

    def create_relationship(
        self,
        source_entity: E,
        target_entity: E,
        description: Optional[str] = None,
        metadata: Optional[dict] = {},
        commit: bool = True,
        db_session: Session = None,
    ) -> R:
        """
        Create a relationship between two entities.
        """
        description_vec = self._get_relationship_description_embedding(
            source_entity.name,
            source_entity.description,
            target_entity.name,
            target_entity.description,
            description,
        )
        relationship = self._relationship_db_model(
            source_entity=source_entity,
            target_entity=target_entity,
            description=description,
            description_vec=description_vec,
            meta=metadata,
            chunk_id=metadata["chunk_id"] if "chunk_id" in metadata else None,
            document_id=metadata["document_id"] if "document_id" in metadata else None,
        )

        with self._session_scope(db_session, commit) as session:
            session.add(relationship)
            return relationship

    def update_relationship(
        self,
        relationship: R,
        update: RelationshipUpdate,
        commit: bool = True,
        db_session: Session = None,
    ) -> R:
        for key, value in update.items():
            if value is None:
                continue
            setattr(relationship, key, value)
            flag_modified(relationship, key)

        # Update embeddings.
        relationship.description_vec = self._get_relationship_description_embedding(
            relationship.source_entity.name,
            relationship.source_entity.description,
            relationship.target_entity.name,
            relationship.target_entity.description,
            relationship.description,
        )

        with self._session_scope(db_session, commit) as session:
            session.add(relationship)
            session.refresh(relationship)
        return relationship

    def delete_relationship(
        self, relationship: R, commit: bool = True, db_session: Session = None
    ):
        with self._session_scope(db_session, commit) as session:
            session.delete(relationship)

    def clear_orphan_entities(self):
        raise NotImplementedError()

    # Relationship Retrieve Operations

    def retrieve_relationships(
        self,
        query: str,
        top_k: int = 10,
        nprobe: Optional[int] = None,
        similarity_threshold: Optional[float] = 0,
    ) -> List[RetrievedRelationship]:
        relationships = self.search_similar_relationships(
            query=QueryBundle(query_str=query),
            top_k=top_k,
            nprobe=nprobe,
            similarity_threshold=similarity_threshold,
        )
        return [
            RetrievedRelationship(
                id=relationship.id,
                source_entity_id=relationship.source_entity_id,
                target_entity_id=relationship.target_entity_id,
                description=relationship.description,
                rag_description=f"{relationship.source_entity.name} -> {relationship.description} -> {relationship.target_entity.name}",
                meta=relationship.meta,
                weight=relationship.weight,
                last_modified_at=relationship.last_modified_at,
                similarity_score=similarity_score,
            )
            for relationship, similarity_score in relationships
        ]

    def search_similar_relationships(
        self,
        query: QueryBundle,
        top_k: int = 10,
        nprobe: Optional[int] = None,
        similarity_threshold: Optional[float] = None,
        distance_range: Optional[Tuple[float, float]] = None,
        weight_threshold: Optional[float] = None,
        exclude_relationship_ids: Optional[List[str]] = None,
        source_entity_ids: Optional[List[int]] = None,
        target_entity_ids: Optional[List[int]] = None,
        metadata_filters: Optional[Dict[str, Any]] = None,
        db_session: Session = None,
    ) -> List[Tuple[R, float]]:
        if query.query_embedding is None:
            query.query_embedding = self._embedding_model.get_query_embedding(
                query.query_str
            )

        nprobe = nprobe or top_k * 10

        subquery = (
            select(
                self._relationship_db_model.id.label("relationship_id"),
                self._relationship_db_model.description_vec.cosine_distance(
                    query.query_embedding
                ).label("embedding_distance"),
            )
            .order_by(asc("embedding_distance"))
            .limit(nprobe)
            .subquery()
        )
        query = (
            select(
                self._relationship_db_model,
                (1 - subquery.c.embedding_distance).label("similarity_score"),
            )
            .join(
                subquery, self._relationship_db_model.id == subquery.c.relationship_id
            )
            .order_by(desc("similarity_score"))
            .limit(top_k)
        )

        if similarity_threshold is not None:
            distance_threshold = 1 - similarity_threshold
            query = query.where(subquery.c.embedding_distance <= distance_threshold)

        if distance_range is not None:
            query = query.where(
                text(
                    "embedding_distance >= :min_distance AND embedding_distance <= :max_distance"
                )
            ).params(min_distance=distance_range[0], max_distance=distance_range[1])

        if weight_threshold is not None:
            query = query.where(subquery.c.similarity_score >= weight_threshold)

        if exclude_relationship_ids is not None and len(exclude_relationship_ids) > 0:
            query = query.where(
                self._relationship_db_model.id.notin_(exclude_relationship_ids)
            )

        if source_entity_ids is not None and len(source_entity_ids) > 0:
            query = query.where(
                self._relationship_db_model.source_entity_id.in_(source_entity_ids)
            )

        if target_entity_ids is not None and len(target_entity_ids) > 0:
            query = query.where(
                self._relationship_db_model.target_entity_id.in_(target_entity_ids)
            )

        if metadata_filters is not None and len(metadata_filters) > 0:
            for key, value in metadata_filters.items():
                json_path = f"$.{key}"
                if isinstance(value, (list, tuple, set)):
                    value_json = json.dumps(list(value))
                    query = query.where(
                        text(f"JSON_CONTAINS(meta->'$.{key}', :value)")
                    ).params(value=value_json)
                else:
                    query = query.where(
                        text("JSON_EXTRACT(meta, :path) = :value")
                    ).params(path=json_path, value=json.dumps(value))

        with self._session_scope(db_session) as session:
            rows = session.exec(query).all()
            return [(row[0], row.similarity_score) for row in rows]

    # Graph Basic Operations

    def list_relationships_by_ids(
        self, relationship_ids: list[int], db_session: Session = None
    ) -> List[R]:
        stmt = (
            select(self._relationship_db_model)
            .options(
                defer(self._relationship_db_model.description_vec),
                joinedload(self._relationship_db_model.source_entity)
                .defer(self._entity_db_model.description_vec)
                .defer(self._entity_db_model.meta_vec),
                joinedload(self._relationship_db_model.target_entity)
                .defer(self._entity_db_model.description_vec)
                .defer(self._entity_db_model.meta_vec),
            )
            .where(self._relationship_db_model.id.in_(relationship_ids))
        )
        with self._session_scope(db_session) as session:
            return session.exec(stmt).all()



import logging
from contextlib import contextmanager
from typing import Any, List, Optional

import sqlalchemy
import tidb_vector
from llama_index.core.vector_stores import MetadataFilters
from sqlalchemy import Engine
from sqlmodel import (
    desc,
    select,
    asc,
    Session,
)
from tidb_vector.sqlalchemy import VectorAdaptor

from autoflow.llms import EmbeddingModel
from autoflow.models.document import Document
from autoflow.storage.doc_store.base import (
    DocumentStore,
    DocumentSearchResult,
    ChunkWithScore,
    D,
    C,
    DocumentSearchMethod,
)


logger = logging.getLogger(__name__)


class TiDBDocumentStore(DocumentStore[D, C]):
    def __init__(
        self,
        db_engine: Engine,
        embedding_model: EmbeddingModel,
        document_db_model: D,
        chunk_db_model: C,
    ) -> None:
        super().__init__()
        self._db_engine = db_engine
        self._embedding_model = embedding_model
        self._document_db_model = document_db_model
        self._chunk_db_model = chunk_db_model

    @classmethod
    def class_name(cls) -> str:
        return "TiDBDocumentStore"

    def ensure_table_schema(self) -> None:
        inspector = sqlalchemy.inspect(self._db_engine)
        existing_table_names = inspector.get_table_names()

        document_model = self._document_db_model
        document_table_name = document_model.__tablename__
        if document_table_name not in existing_table_names:
            document_model.metadata.create_all(
                self._db_engine, tables=[document_model.__table__]
            )
            logger.info(
                f"Document table <{document_table_name}> has been created successfully."
            )
        else:
            logger.info(
                f"Document table <{document_table_name}> is already exists, no action to do."
            )

        chunk_model = self._chunk_db_model
        chunk_table_name = chunk_model.__tablename__
        if chunk_table_name not in existing_table_names:
            chunk_model.metadata.create_all(
                self._db_engine, tables=[chunk_model.__table__]
            )
            VectorAdaptor(self._db_engine).create_vector_index(
                chunk_model.text_vec, tidb_vector.DistanceMetric.COSINE
            )
            logger.info(
                f"Chunk table <{chunk_table_name}> has been created successfully."
            )
        else:
            logger.info(
                f"Chunk table <{chunk_table_name}> is already exists, no action to do."
            )

    def drop_table_schema(self) -> None:
        inspector = sqlalchemy.inspect(self._db_engine)
        existed_table_names = inspector.get_table_names()

        document_model = self._document_db_model
        document_table_name = document_model.__tablename__
        if document_table_name in existed_table_names:
            document_model.metadata.drop_all(
                self._db_engine, tables=[document_model.__table__]
            )
            logger.info(
                f"Document table <{document_table_name}> has been dropped successfully."
            )
        else:
            logger.info(
                f"Document table <{document_table_name}> is not exists, no action to do."
            )

        chunk_model = self._chunk_db_model
        chunk_table_name = chunk_model.__tablename__
        if chunk_table_name in existed_table_names:
            chunk_model.metadata.drop_all(
                self._db_engine, tables=[chunk_model.__table__]
            )
            logger.info(
                f"Chunk table <{chunk_table_name}> has been dropped successfully."
            )
        else:
            logger.info(
                f"Chunk table <{chunk_table_name}> is not exists, no action to do."
            )

    @contextmanager
    def _session_scope(self, session: Optional[Session] = None, commit: bool = False):
        """Provide a transactional scope around a series of operations."""
        should_close = session is None
        session = session or Session(self._db_engine)

        try:
            yield session
            if commit:
                session.commit()
            else:
                session.flush()
        except Exception as e:
            session.rollback()
            raise e
        finally:
            if should_close:
                session.close()

    def add(self, documents: List[Document], **add_kwargs: Any) -> List[Document]:
        with Session(self._db_engine) as db_session:
            for doc in documents:
                db_session.merge(doc)
                db_session.add(doc)
            db_session.commit()
            for doc in documents:
                db_session.refresh(doc)
            return documents

    def delete(self, document_id: int) -> None:
        with Session(self._db_engine) as db_session:
            doc = db_session.get(self._document_db_model, document_id)
            if doc is None:
                raise ValueError("Document with id #{} not found".format(document_id))
            # TODO: Delete the chunks associated with the document.
            db_session.delete(doc)
            db_session.commit()

    def list(self) -> List[D]:
        with Session(self._db_engine) as db_session:
            query = select(self._document_db_model)
            return db_session.exec(query).all()

    def get(self, document_id: int) -> D:
        with Session(self._db_engine) as db_session:
            doc = db_session.get(self._document_db_model, document_id)
            if doc is None:
                raise ValueError("Document with id #{} not found".format(document_id))
            return doc

    def add_doc_chunks(self, chunks: List[C]) -> List[C]:
        chunk_texts = [chunk.text for chunk in chunks]
        embeddings = self._embedding_model.get_text_embedding_batch(chunk_texts)
        for chunk, embedding in zip(chunks, embeddings):
            chunk.text_vec = embedding

        with Session(self._db_engine) as db_session:
            db_session.bulk_save_objects(chunks)
            db_session.commit()
            return chunks

    # TODO: call the low-level database API.
    def search(
        self,
        query: str,
        search_method: List[DocumentSearchMethod] = [
            DocumentSearchMethod.VECTOR_SEARCH
        ],
        top_k: Optional[int] = None,
        similarity_threshold: Optional[float] = None,
        similarity_nprobe: Optional[int] = None,
        similarity_top_k: Optional[int] = 5,
        **kwargs: Any,
    ) -> DocumentSearchResult:
        # TODO: Support Hybrid search.
        with self._session_scope() as db_session:
            chunks_with_score = self._vector_search(
                query=query,
                # metadata_filters=query.metadata_filters,
                nprobe=similarity_nprobe,
                similarity_top_k=similarity_top_k,
                db_session=db_session,
            )
            # chunks_with_score = self._rerank_chunks(chunks_with_score)
            document_ids = {c.chunk.document_id for c in chunks_with_score}
            documents = self._list_documents_by_ids(
                list(document_ids), db_session=db_session
            )
            return DocumentSearchResult(
                chunks=chunks_with_score,
                documents=documents,
            )

    def _vector_search(
        self,
        query: str,
        # metadata_filters: Optional[MetadataFilters] = None,
        nprobe: Optional[int] = None,
        similarity_top_k: int = 5,
        db_session: Optional[Session] = None,
    ) -> List[ChunkWithScore]:
        query_embedding = self._embedding_model.get_query_embedding(query)
        nprobe = nprobe if nprobe else similarity_top_k * 10

        # Base query for vector similarity
        subquery = (
            select(
                self._chunk_db_model.id.label("chunk_id"),
                self._chunk_db_model.text_vec.cosine_distance(query_embedding).label(
                    "embedding_distance"
                ),
            )
            .order_by(asc("embedding_distance"))
            .limit(nprobe)
            .subquery()
        )

        # Main query with metadata filters
        query = select(
            self._chunk_db_model,
            (1 - subquery.c.embedding_distance).label("similarity_score"),
        ).join(subquery, self._chunk_db_model.id == subquery.c.chunk_id)

        # Apply metadata filters if provided
        # TODO: Implement metadata filters.

        # Apply final ordering and limit
        query = query.order_by(desc("similarity_score")).limit(similarity_top_k)

        with self._session_scope(db_session) as db_session:
            results = db_session.exec(query)
            return [
                ChunkWithScore(chunk=chunk, score=score) for chunk, score in results
            ]

    def _fulltext_search(
        self,
        query_str: str,
        metadata_filters: Optional[MetadataFilters] = None,
        top_k: Optional[int] = 5,
    ) -> List[ChunkWithScore]:
        raise NotImplementedError()

    def _rerank_chunks(
        self, chunks_with_score: List[ChunkWithScore]
    ) -> List[ChunkWithScore]:
        raise NotImplementedError("Reranking is not implemented for TiDBDocumentStore.")

    def _list_documents_by_ids(
        self, document_ids: List[int], db_session: Session = None
    ) -> List[D]:
        with self._session_scope(db_session) as db_session:
            query = select(self._document_db_model).where(
                self._document_db_model.id.in_(document_ids)
            )
            return db_session.exec(query).all()



from pydantic import PrivateAttr
from sqlalchemy.orm import registry
from sqlmodel import SQLModel, Field

default_registry = registry()

Base = default_registry.generate_base()


class TiDBModel(SQLModel):
    # Generated on query
    _similarity_score: float = PrivateAttr(None)
    _score: float = PrivateAttr(None)

    @property
    def similarity_score(self) -> float:
        return self._similarity_score

    @similarity_score.setter
    def similarity_score(self, val: float):
        self._similarity_score = val

    @property
    def score(self) -> float:
        return self._score

    @score.setter
    def score(self, val: float):
        self._score = val


Field = Field



from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum

from typing import (
    List,
    Optional,
    TypeVar,
    Type,
    Generic,
    Mapping,
    Any,
    Tuple,
    Sequence,
    Collection,
    Dict,
)
from uuid import UUID

from pydantic import BaseModel, model_validator, Field
from sqlmodel import SQLModel, Session

from autoflow.indices.knowledge_graph.schema import AIKnowledgeGraph
from autoflow.llms import EmbeddingModel
from autoflow.models.entity import EntityType
from autoflow.storage.schema import QueryBundle


# Entity


class EntityCreate(BaseModel):
    entity_type: Optional[EntityType] = EntityType.original
    name: Optional[str] = None
    description: Optional[str] = None
    meta: Optional[dict] = None


class SynopsisEntityCreate(EntityCreate):
    topic: str
    entities: List[int] = Field(description="The id list of the related entities")

    @model_validator(mode="after")
    def validate_entities(self):
        if len(self.entities) == 0:
            raise ValueError("Entities list should not be empty")
        return self


class EntityFilters(BaseModel):
    entity_ids: Optional[List[int]] = None
    entity_type: Optional[EntityType] = None
    search: Optional[str] = None


class EntityUpdate(BaseModel):
    description: Optional[str] = None
    meta: Optional[dict] = None


class EntityDegree(BaseModel):
    out_degree: int = 0
    in_degree: int = 0
    degrees: int = 0


# Relationship


class RelationshipCreate(BaseModel):
    source_entity_id: int
    target_entity_id: int
    description: str


class RelationshipUpdate(BaseModel):
    description: Optional[str] = None


class RelationshipFilters(BaseModel):
    target_entity_id: Optional[int] = None
    source_entity_id: Optional[int] = None
    chunk_ids: Optional[List[UUID]] = None
    relationship_ids: Optional[List[int]] = None
    search: Optional[str] = None


# Stored Knowledge Graph


class StoredKnowledgeGraphVersion(int, Enum):
    V1 = 1


class StoredSubGraph(BaseModel):
    query: Optional[str] = None
    knowledge_base_id: Optional[int] = None
    entities: Optional[list[int]] = None
    relationships: Optional[list[int]] = None


class StoredKnowledgeGraph(StoredSubGraph):
    """
    StoredKnowledgeGraph represents the structure of the knowledge graph stored in the database.
    """

    # If not provided, it means that the old version of the storage format is used, which only
    # stores entities and relationships information.
    version: Optional[int] = StoredKnowledgeGraphVersion.V1
    knowledge_base_ids: Optional[list[int]] = []
    subgraphs: Optional[list["StoredSubGraph"]] = None


# Retrieved Knowledge Graph


class RetrievedEntity(BaseModel):
    id: int = Field(description="ID of the entity")
    knowledge_base_id: Optional[int] = Field(
        description="ID of the knowledge base", default=None
    )
    entity_type: Optional[EntityType] = Field(
        description="Type of the entity", default=EntityType.original
    )
    name: str = Field(description="Name of the entity")
    description: str = Field(description="Description of the entity")
    meta: Optional[Mapping[str, Any]] = Field(description="Metadata of the entity")
    similarity_score: Optional[float] = Field(default=None)

    @property
    def global_id(self) -> str:
        return f"{self.knowledge_base_id or 0}-{self.id}"

    def __hash__(self):
        return hash(self.global_id)


class RetrievedRelationship(BaseModel):
    id: int = Field(description="ID of the relationship")
    knowledge_base_id: int = Field(description="ID of the knowledge base", default=None)
    source_entity_id: int = Field(description="ID of the source entity")
    target_entity_id: int = Field(description="ID of the target entity")
    description: str = Field(description="Description of the relationship")
    meta: Optional[Mapping[str, Any]] = Field(
        description="Metadata of the relationship"
    )
    rag_description: Optional[str] = Field(
        description="RAG description of the relationship"
    )
    weight: Optional[float] = Field(description="Weight of the relationship")
    last_modified_at: Optional[datetime] = Field(
        description="Last modified at of the relationship", default=None
    )
    similarity_score: Optional[float] = Field()

    @property
    def global_id(self) -> str:
        return f"{self.knowledge_base_id or 0}-{self.id}"

    def __hash__(self):
        return hash(self.global_id)


class RetrievedSubGraph(BaseModel):
    query: Optional[str | list[str]] = Field(
        description="List of queries that are used to retrieve the knowledge graph",
        default=None,
    )
    # knowledge_base: Optional[KnowledgeBaseDescriptor] = Field(
    #     description="The knowledge base that the knowledge graph is retrieved from",
    #     default=None,
    # )
    entities: List[RetrievedEntity] = Field(
        description="List of entities in the knowledge graph", default_factory=list
    )
    relationships: List[RetrievedRelationship] = Field(
        description="List of relationships in the knowledge graph", default_factory=list
    )


class RetrievedKnowledgeGraph(RetrievedSubGraph):
    """
    RetrievedKnowledgeGraph represents the structure of the knowledge graph retrieved
    from the knowledge base.
    """

    # knowledge_bases: Optional[List[KnowledgeBaseDescriptor]] = Field(
    #     description="List of knowledge bases that the knowledge graph is retrieved from",
    #     default_factory=list,
    # )

    # subgraphs: Optional[List["RetrievedSubGraph"]] = Field(
    #     description="List of subgraphs of the knowledge graph", default_factory=list
    # )

    def to_subqueries_dict(self) -> dict:
        """
        For forward compatibility, we need to convert the subgraphs to a dictionary
        of subqueries and then pass it to the prompt template.
        """
        subqueries = {}
        for subgraph in self.subgraphs:
            if subgraph.query not in subqueries:
                subqueries[subgraph.query] = {
                    "entities": [e.model_dump() for e in subgraph.entities],
                    "relationships": [r.model_dump() for r in subgraph.relationships],
                }
            else:
                subqueries[subgraph.query]["entities"].extend(
                    [e.model_dump() for e in subgraph.entities]
                )
                subqueries[subgraph.query]["relationships"].extend(
                    [r.model_dump() for r in subgraph.relationships]
                )

        return subqueries

    def to_stored_graph_dict(self) -> dict:
        subgraph = self.to_stored_graph()
        return subgraph.model_dump()

    def to_stored_graph(self) -> StoredKnowledgeGraph:
        return StoredKnowledgeGraph(
            query=self.query,
            knowledge_base_id=self.knowledge_base.id if self.knowledge_base else None,
            knowledge_base_ids=[kb.id for kb in self.knowledge_bases]
            if self.knowledge_bases
            else None,
            entities=[e.id for e in self.entities],
            relationships=[r.id for r in self.relationships],
            subgraphs=[s.to_stored_graph() for s in self.subgraphs],
        )


KnowledgeGraphRetrievalResult = RetrievedKnowledgeGraph

# Graph Search


class GraphSearchAlgorithm(str, Enum):
    WEIGHTED_SEARCH = "weighted"


# Knowledge Graph Store

E = TypeVar("E", SQLModel, Type[SQLModel])
R = TypeVar("R", SQLModel, Type[SQLModel])
C = TypeVar("C", SQLModel, Type[SQLModel])


class KnowledgeGraphStore(ABC, Generic[E, R, C]):
    """Abstract base class for knowledge graph storage"""

    def __init__(self, embed_model: EmbeddingModel):
        self._embed_model = embed_model

    # Schema Operations
    @abstractmethod
    def ensure_table_schema(self) -> None:
        """Ensure database table schema exists"""
        raise NotImplementedError

    @abstractmethod
    def drop_table_schema(self) -> None:
        """Drop database table schema"""
        raise NotImplementedError

    @abstractmethod
    def save_knowledge_graph(
        self, knowledge_graph: AIKnowledgeGraph, chunk: SQLModel
    ) -> None:
        raise NotImplementedError

    # Entity Basic Operations

    @abstractmethod
    def list_entities(
        self,
        filters: Optional[EntityFilters] = EntityFilters(),
        db_session: Session = None,
    ) -> Sequence[E]:
        """List all entities matching the filters"""
        raise NotImplementedError

    @abstractmethod
    def get_entity_by_id(self, entity_id: int) -> Type[E]:
        """Get entity by ID"""
        raise NotImplementedError

    @abstractmethod
    def must_get_entity_by_id(self, entity_id: int) -> Type[E]:
        """Get entity by ID, raise error if not found"""
        raise NotImplementedError

    @abstractmethod
    def create_entity(
        self, create: EntityCreate, commit: bool = True, db_session: Session = None
    ) -> E:
        """Create a new entity"""
        raise NotImplementedError

    @abstractmethod
    def find_or_create_entity(
        self, create: EntityCreate, commit: bool = True, db_session: Session = None
    ) -> E:
        """Find existing entity or create new one"""
        raise NotImplementedError

    @abstractmethod
    def update_entity(
        self,
        entity: Type[E],
        update: EntityUpdate,
        commit: bool = True,
        db_session: Session = None,
    ) -> Type[E]:
        """Update an existing entity"""
        raise NotImplementedError

    @abstractmethod
    def delete_entity(
        self, entity: Type[E], commit: bool = True, db_session: Session = None
    ) -> None:
        """Delete an entity"""
        raise NotImplementedError

    @abstractmethod
    def list_entity_relationships(self, entity_id: int) -> List[R]:
        """List all relationships for an entity"""
        raise NotImplementedError

    # Entity Degree Operations
    @abstractmethod
    def calc_entity_out_degree(self, entity_id: int) -> Optional[int]:
        """Calculate out-degree of an entity"""
        raise NotImplementedError

    @abstractmethod
    def calc_entity_in_degree(self, entity_id: int) -> Optional[int]:
        """Calculate in-degree of an entity"""
        raise NotImplementedError

    @abstractmethod
    def calc_entity_degree(self, entity_id: int) -> Optional[int]:
        """Calculate total degree of an entity"""
        raise NotImplementedError

    @abstractmethod
    def bulk_calc_entities_degrees(
        self, entity_ids: Collection[int]
    ) -> Mapping[int, EntityDegree]:
        """Calculate degrees for multiple entities in bulk"""
        raise NotImplementedError

    # Entity Retrieve Operations
    @abstractmethod
    def retrieve_entities(
        self,
        query: str,
        entity_type: EntityType = EntityType.original,
        top_k: int = 10,
        nprobe: Optional[int] = None,
        similarity_threshold: Optional[float] = None,
    ) -> List[RetrievedEntity]:
        """Retrieve entities based on semantic search"""
        raise NotImplementedError

    @abstractmethod
    def search_similar_entities(
        self,
        query: QueryBundle,
        top_k: int = 10,
        nprobe: Optional[int] = None,
        entity_type: EntityType = EntityType.original,
        similarity_threshold: Optional[float] = None,
    ) -> List[Tuple[E, float]]:
        """Search for similar entities using vector similarity"""
        raise NotImplementedError

    # Relationship Basic Operations
    @abstractmethod
    def get_relationship_by_id(self, relationship_id: int) -> R:
        """Get relationship by ID"""
        raise NotImplementedError

    @abstractmethod
    def list_relationships(self, filters: RelationshipFilters) -> Sequence[R]:
        """List all relationships matching the filters"""
        raise NotImplementedError

    @abstractmethod
    def create_relationship(
        self,
        source_entity: E,
        target_entity: E,
        description: Optional[str] = None,
        metadata: Optional[dict] = {},
        commit: bool = True,
        db_session: Session = None,
    ) -> R:
        """Create a new relationship between entities"""
        raise NotImplementedError

    @abstractmethod
    def update_relationship(
        self,
        relationship: R,
        update: RelationshipUpdate,
        commit: bool = True,
        db_session: Session = None,
    ) -> R:
        """Update an existing relationship"""
        raise NotImplementedError

    @abstractmethod
    def delete_relationship(
        self, relationship: R, commit: bool = True, db_session: Session = None
    ):
        """Delete a relationship"""
        raise NotImplementedError

    @abstractmethod
    def clear_orphan_entities(self, db_session: Session = None):
        """Remove entities that have no relationships"""
        raise NotImplementedError

    # Relationship Retrieve Operations
    @abstractmethod
    def retrieve_relationships(
        self,
        query: str,
        top_k: int = 10,
        nprobe: Optional[int] = None,
        similarity_threshold: Optional[float] = 0,
    ) -> List[RetrievedRelationship]:
        """Retrieve relationships based on semantic search"""
        raise NotImplementedError

    @abstractmethod
    def search_similar_relationships(
        self,
        query: QueryBundle,
        top_k: int = 10,
        nprobe: Optional[int] = None,
        similarity_threshold: Optional[float] = None,
        distance_range: Optional[Tuple[float, float]] = None,
        weight_threshold: Optional[float] = None,
        exclude_relationship_ids: Optional[List[str]] = None,
        source_entity_ids: Optional[List[int]] = None,
        target_entity_ids: Optional[List[int]] = None,
        metadata_filters: Optional[Dict[str, Any]] = None,
    ) -> List[Tuple[R, float]]:
        """Search for similar relationships using vector similarity

        Args:
            query: Query bundle containing search text or embedding
            top_k: Number of results to return
            nprobe: Number of clusters to probe in ANN search
            similarity_threshold: Minimum similarity score threshold
            distance_range: Range of acceptable distances
            weight_threshold: Minimum relationship weight threshold
            exclude_relationship_ids: Relationship IDs to exclude
            source_entity_ids: Filter by source entity IDs
            target_entity_ids: Filter by target entity IDs
            metadata_filters: Filter by metadata fields

        Returns:
            List of tuples containing relationship and similarity score
        """
        raise NotImplementedError

    # Graph Basic Operations
    @abstractmethod
    def list_relationships_by_ids(
        self, relationship_ids: list[int], **kwargs
    ) -> List[R]:
        """Get multiple relationships by their IDs"""
        raise NotImplementedError

    # Knowledge Graph Retrieve Operations
    def search(
        self,
        query: QueryBundle,
        depth: int = 2,
        include_meta: bool = False,
        metadata_filters: Optional[dict] = None,
        search_algorithm: Optional[
            GraphSearchAlgorithm
        ] = GraphSearchAlgorithm.WEIGHTED_SEARCH,
        **kwargs,
    ) -> RetrievedKnowledgeGraph:
        """Search the knowledge graph using configurable search algorithm

        Args:
            query: Query bundle containing search text or embedding
            depth: Maximum search depth in the graph
            include_meta: Whether to include metadata in results
            metadata_filters: Filters to apply on metadata
            search_algorithm: Algorithm class to use for graph search

        Returns:
            Retrieved subgraph containing matching entities and relationships

        """
        # Ensure query has embedding
        if query.query_embedding is None and hasattr(self, "_embed_model"):
            query.query_embedding = self._embed_model.get_query_embedding(
                query.query_str
            )

        # Initialize and execute search algorithm
        if search_algorithm is GraphSearchAlgorithm.WEIGHTED_SEARCH:
            from autoflow.storage.graph_store.algorithms.weighted import (
                WeightedGraphSearchRetriever,
            )

            retriever = WeightedGraphSearchRetriever(self, **kwargs)
        else:
            raise NotImplementedError(f"Unknown search algorithm <{search_algorithm}>")

        relationships, entities = retriever.search(
            query_embedding=query.query_embedding,
            depth=depth,
            metadata_filters=metadata_filters,
        )

        # Construct result graph
        return RetrievedKnowledgeGraph(
            relationships=[
                RetrievedRelationship(
                    id=r.id,
                    source_entity_id=r.relationship.source_entity_id,
                    target_entity_id=r.relationship.target_entity_id,
                    description=r.relationship.description,
                    rag_description=f"{r.relationship.source_entity.name} -> {r.relationship.description} -> {r.relationship.target_entity.name}",
                    meta=r.relationship.meta if include_meta else None,
                    weight=r.relationship.weight,
                    last_modified_at=r.relationship.last_modified_at
                    if hasattr(r.relationship, "last_modified_at")
                    else None,
                    similarity_score=r.score if hasattr(r, "score") else None,
                )
                for r in relationships
            ],
            entities=[
                RetrievedEntity(
                    id=e.id,
                    entity_type=e.entity_type,
                    name=e.name,
                    description=e.description,
                    meta=e.meta if include_meta else None,
                )
                for e in entities
            ],
        )



import re
from typing import Any, Dict, List, Optional, Sequence, Callable

from llama_index.core.callbacks.base import CallbackManager
from llama_index.core.node_parser import SentenceSplitter
from llama_index.core.node_parser.interface import NodeParser
from llama_index.core.node_parser.node_utils import build_nodes_from_splits
from llama_index.core.schema import BaseNode, MetadataMode, TextNode
from llama_index.core.utils import get_tqdm_iterable, get_tokenizer
from llama_index.core.bridge.pydantic import Field, PrivateAttr


DEFAULT_CHUNK_HEADER_LEVEL = 2
DEFAULT_CHUNK_SIZE = 1200


class MarkdownNodeParser(NodeParser):
    """Markdown node parser.
    Splits a document into Nodes using custom Markdown splitting logic.
    Args:
        include_metadata (bool): whether to include metadata in nodes
        include_prev_next_rel (bool): whether to include prev/next relationships
    """

    chunk_size: int = Field(
        default=DEFAULT_CHUNK_SIZE,
        description="The token chunk size for each chunk.",
        gt=0,
    )
    chunk_header_level: int = Field(
        default=DEFAULT_CHUNK_HEADER_LEVEL,
        description="The header level to split on",
        ge=1,
        le=6,
    )
    _tokenizer: Callable = PrivateAttr()

    def __init__(
        self,
        chunk_header_level: int = DEFAULT_CHUNK_HEADER_LEVEL,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        tokenizer: Optional[Callable] = None,
        include_metadata: bool = True,
        include_prev_next_rel: bool = True,
        callback_manager: Optional[CallbackManager] = None,
    ):
        super().__init__(
            chunk_header_level=chunk_header_level,
            chunk_size=chunk_size,
            include_metadata=include_metadata,
            include_prev_next_rel=include_prev_next_rel,
            callback_manager=callback_manager,
        )
        self.callback_manager = callback_manager or CallbackManager([])
        self._tokenizer = tokenizer or get_tokenizer()

    @classmethod
    def from_defaults(
        cls,
        chunk_header_level: int = DEFAULT_CHUNK_HEADER_LEVEL,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        tokenizer: Optional[Callable] = None,
        include_metadata: bool = True,
        include_prev_next_rel: bool = True,
        callback_manager: Optional[CallbackManager] = None,
    ) -> "MarkdownNodeParser":
        callback_manager = callback_manager or CallbackManager([])
        tokenizer = tokenizer or get_tokenizer()
        return cls(
            chunk_header_level=chunk_header_level,
            chunk_size=chunk_size,
            tokenizer=tokenizer,
            include_metadata=include_metadata,
            include_prev_next_rel=include_prev_next_rel,
            callback_manager=callback_manager,
        )

    @classmethod
    def class_name(cls) -> str:
        """Get class name."""
        return "MarkdownNodeParser"

    def _parse_nodes(
        self,
        nodes: Sequence[BaseNode],
        show_progress: bool = False,
        **kwargs: Any,
    ) -> List[BaseNode]:
        all_nodes: List[BaseNode] = []
        nodes_with_progress = get_tqdm_iterable(nodes, show_progress, "Parsing nodes")

        for node in nodes_with_progress:
            splitted_nodes = self.get_nodes_from_node(
                node,
                self.chunk_header_level,
                self.chunk_size * 0.7,
                self.chunk_size * 1.1,
            )
            nodes = []
            for sn in splitted_nodes:
                header_level = sn.metadata.get("Header_Level")
                if header_level:
                    for _hl in range(1, header_level + 1)[::-1]:
                        if (
                            f"Header_{_hl}" in sn.metadata
                            and sn.metadata[f"Header_{_hl}"] not in sn.text
                        ):
                            sn.text = (
                                f'{"#" * _hl} {sn.metadata[f"Header_{_hl}"]}\n\n'
                                + sn.text
                            )
                n = build_nodes_from_splits([sn.text], node, id_func=self.id_func)[0]
                if self.include_metadata:
                    n.metadata = {**node.metadata, **sn.metadata}
                nodes.append(n)
            all_nodes.extend(nodes)

        return all_nodes

    def get_nodes_from_node(
        self,
        node: BaseNode,
        chunk_header_level: int,
        chunk_size_small_threshold: float,
        chunk_size_large_threshold: float,
    ) -> List[TextNode]:
        # print(chunk_header_level, chunk_size_small_threshold, chunk_size_large_threshold)
        """Get nodes from document."""
        text = node.get_content(metadata_mode=MetadataMode.NONE)
        markdown_nodes = []
        lines = text.split("\n")
        metadata: Dict[str, str] = node.metadata
        code_block = False
        current_section = ""
        first_header = True

        for line in lines:
            if line.lstrip().startswith("```"):
                code_block = not code_block
            header_match = re.match(r"^(#+)\s(.*)", line)
            if header_match and not code_block:
                current_header_level = len(header_match.group(1).strip())
                if current_section != "" and current_header_level == chunk_header_level:
                    if first_header:
                        # skip the first header, merge it with the first section (usually the title of the document)
                        first_header = False
                    else:
                        markdown_nodes.append(
                            self._build_node_from_split(
                                current_section.strip(), node, metadata
                            )
                        )
                        current_section = ""
                if current_header_level <= chunk_header_level:
                    metadata = self._update_metadata(
                        metadata, header_match.group(2), current_header_level
                    )
                current_section += line + "\n"
            else:
                current_section += line + "\n"

        markdown_nodes.append(
            self._build_node_from_split(current_section.strip(), node, metadata)
        )
        return self._normalize_node_sizes(
            markdown_nodes, chunk_size_small_threshold, chunk_size_large_threshold
        )

    def _normalize_node_sizes(
        self,
        nodes: List[TextNode],
        chunk_size_small_threshold: float,
        chunk_size_large_threshold: float,
    ) -> List[TextNode]:
        # 1. Split the big node into multiple small nodes
        # 2. Merge the small nodes into a big node if they are too small
        # 3. Make all the nodes as much as possible close to the chunk size
        nodes_token_size = [self._token_size(node.text) for node in nodes]
        normalized_nodes = []
        buffer = []
        node_count = len(nodes)
        i = 0

        while i < node_count:
            node = nodes[i]
            this_chunk_size = nodes_token_size[i]
            if this_chunk_size < chunk_size_small_threshold:
                # if the last node is too small, merge it with the previous one
                if (
                    not buffer
                    and i == (node_count - 1)
                    and i > 0
                    and nodes_token_size[i - 1] + this_chunk_size
                    < chunk_size_large_threshold
                ):
                    normalized_nodes[-1].text += "\n\n" + node.text
                    i += 1
                    continue
                buffer.append(this_chunk_size)
                total = sum(buffer)
                while (
                    (i + 1) < node_count
                    and nodes_token_size[i + 1] < self.chunk_size
                    and total + nodes_token_size[i + 1] <= chunk_size_large_threshold
                ):
                    i += 1
                    buffer.append(nodes_token_size[i])
                    total += nodes_token_size[i]
                # output the sum of the buffer
                buffer_nodes = nodes[i - len(buffer) + 1 : i + 1]
                normalized_nodes.append(
                    TextNode(
                        text="\n\n".join([node.text for node in buffer_nodes]),
                        metadata=buffer_nodes[0].metadata,
                    )
                )
                i += 1
                buffer.clear()
            elif this_chunk_size > chunk_size_large_threshold:
                # split into multiple nodes with next header level and bigger chunk size
                md_splitted_nodes = self.get_nodes_from_node(
                    node,
                    self.chunk_header_level + 1,
                    chunk_size_small_threshold,
                    chunk_size_large_threshold * 1.1,
                )
                for n in md_splitted_nodes:
                    _chunk_size = self._token_size(n.text)
                    if _chunk_size > chunk_size_large_threshold * 1.1:
                        # using sentence splitter to split the node if it's still too large
                        sentence_splitted_nodes = SentenceSplitter(
                            chunk_size=int(chunk_size_large_threshold), separator="\n\n"
                        ).get_nodes_from_documents([n])
                        normalized_nodes.extend(sentence_splitted_nodes)
                    else:
                        normalized_nodes.append(n)
                i += 1
            else:
                normalized_nodes.append(node)
                i += 1
        return normalized_nodes

    def _update_metadata(
        self, headers_metadata: dict, new_header: str, new_header_level: int
    ) -> dict:
        """Update the markdown headers for metadata.
        Removes all headers that are equal or less than the level
        of the newly found header
        """
        updated_headers = {}

        for i in range(1, new_header_level):
            key = f"Header_{i}"
            if key in headers_metadata:
                updated_headers[key] = headers_metadata[key]

        updated_headers[f"Header_{new_header_level}"] = new_header
        updated_headers["Header_Level"] = new_header_level
        return updated_headers

    def _build_node_from_split(
        self,
        text_split: str,
        node: BaseNode,
        metadata: dict,
    ) -> TextNode:
        """Build node from single text split."""
        node = build_nodes_from_splits([text_split], node, id_func=self.id_func)[0]

        if self.include_metadata:
            node.metadata = {**node.metadata, **metadata}

        return node

    def _token_size(self, text: str) -> int:
        return len(self._tokenizer(text))



import enum
from typing import Union, Dict
from pydantic import BaseModel
from sqlmodel import Field
from autoflow.datasources.mime_types import SupportedMimeTypes
from autoflow.llms import EmbeddingModelConfig, ChatModelConfig
from autoflow.schema import IndexMethod

# Index Methods

DEFAULT_INDEX_METHODS = [IndexMethod.VECTOR_SEARCH]

# Chunking Settings

DEFAULT_CHUNK_SIZE = 1200
DEFAULT_CHUNK_HEADER_LEVEL = 2
SENTENCE_CHUNK_OVERLAP = 200
DEFAULT_PARAGRAPH_SEP = "\n\n\n"


class ChunkSplitter(str, enum.Enum):
    SENTENCE_SPLITTER = "SentenceSplitter"
    MARKDOWN_NODE_PARSER = "MarkdownNodeParser"


class SentenceSplitterOptions(BaseModel):
    chunk_size: int = Field(
        description="The token chunk size for each chunk.",
        default=1000,
        gt=0,
    )
    chunk_overlap: int = Field(
        description="The overlap size for each chunk.",
        default=SENTENCE_CHUNK_OVERLAP,
        gt=0,
    )
    paragraph_separator: str = Field(
        description="The paragraph separator for splitting the text.",
        default=DEFAULT_PARAGRAPH_SEP,
    )


class MarkdownNodeParserOptions(BaseModel):
    chunk_size: int = Field(
        description="The token chunk size for each chunk.",
        default=1000,
        gt=0,
    )
    chunk_header_level: int = Field(
        description="The header level to split on",
        default=DEFAULT_CHUNK_HEADER_LEVEL,
        ge=1,
        le=6,
    )


class ChunkSplitterConfig(BaseModel):
    splitter: ChunkSplitter = Field(default=ChunkSplitter.SENTENCE_SPLITTER)
    splitter_options: Union[SentenceSplitterOptions, MarkdownNodeParserOptions] = (
        Field()
    )


class ChunkingMode(str, enum.Enum):
    GENERAL = "general"
    ADVANCED = "advanced"


class BaseChunkingConfig(BaseModel):
    mode: ChunkingMode = Field(default=ChunkingMode.GENERAL)


class GeneralChunkingConfig(BaseChunkingConfig):
    mode: ChunkingMode = Field(default=ChunkingMode.GENERAL)
    chunk_size: int = Field(default=DEFAULT_CHUNK_SIZE, gt=0)
    chunk_overlap: int = Field(default=SENTENCE_CHUNK_OVERLAP, gt=0)
    paragraph_separator: str = Field(default=DEFAULT_PARAGRAPH_SEP)


class AdvancedChunkingConfig(BaseChunkingConfig):
    mode: ChunkingMode = Field(default=ChunkingMode.ADVANCED)
    rules: Dict[SupportedMimeTypes, ChunkSplitterConfig] = Field(default_factory=list)


ChunkingConfig = Union[GeneralChunkingConfig | AdvancedChunkingConfig]


# Knowledge Base Config.


class KnowledgeBaseConfig(BaseModel):
    chat_model: ChatModelConfig = None
    embedding_model: EmbeddingModelConfig = None
    chunking_config: ChunkingConfig = GeneralChunkingConfig()



from typing import Any, List, Optional
from llama_index.core.bridge.pydantic import Field
from llama_index.core.embeddings import BaseEmbedding


def get_embeddings(
    api_key: str,
    api_base: str,
    model_name: str,
    input: List[str],
    timeout: int = 60,
    **kwargs: Any,
) -> List[List[float]]:
    """
    Retrieve embeddings for a given list of input strings using the specified model.

    Args:
        api_key (str): The API key for authentication.
        api_base (str): The base URL of the LiteLLM proxy server.
        model_name (str): The name of the model to use for generating embeddings.
        input (List[str]): A list of input strings for which embeddings are to be generated.
        timeout (float): The timeout value for the API call, default 60 secs.
        **kwargs (Any): Additional keyword arguments to be passed to the embedding function.

    Returns:
        List[List[float]]: A list of embeddings, where each embedding corresponds to an input string.
    """
    from litellm import embedding

    response = embedding(
        api_key=api_key,
        api_base=api_base,
        model=model_name,
        input=input,
        timeout=timeout,
        **kwargs,
    )
    return [result["embedding"] for result in response.data]


class LiteLLMEmbedding(BaseEmbedding):
    model_name: str = Field(description="The name of the embedding model.")
    api_key: Optional[str] = Field(
        default=None,
        description="OpenAI key. If not provided, the proxy server must be configured with the key.",
    )
    api_base: Optional[str] = Field(
        default=None, description="The base URL of the LiteLLM proxy."
    )
    dimensions: Optional[int] = Field(
        default=None,
        description=(
            "The number of dimensions the resulting output embeddings should have. "
            "Only supported in text-embedding-3 and later models."
        ),
    )
    timeout: Optional[int] = Field(
        default=60, description="Timeout for each request.", ge=0
    )

    @classmethod
    def class_name(cls) -> str:
        return "lite-llm"

    async def _aget_query_embedding(self, query: str) -> List[float]:
        return self._get_query_embedding(query)

    async def _aget_text_embedding(self, text: str) -> List[float]:
        return self._get_text_embedding(text)

    def _get_query_embedding(self, query: str) -> List[float]:
        embeddings = get_embeddings(
            api_key=self.api_key,
            api_base=self.api_base,
            model_name=self.model_name,
            dimensions=self.dimensions,
            timeout=self.timeout,
            input=[query],
        )
        return embeddings[0]

    def _get_text_embedding(self, text: str) -> List[float]:
        embeddings = get_embeddings(
            api_key=self.api_key,
            api_base=self.api_base,
            model_name=self.model_name,
            dimensions=self.dimensions,
            timeout=self.timeout,
            input=[text],
        )
        return embeddings[0]

    def _get_text_embeddings(self, texts: List[str]) -> List[List[float]]:
        return get_embeddings(
            api_key=self.api_key,
            api_base=self.api_base,
            model_name=self.model_name,
            dimensions=self.dimensions,
            timeout=self.timeout,
            input=texts,
        )



import uuid
from pathlib import Path
from typing import List, Optional, Dict, Any

from llama_index.core.node_parser import SentenceSplitter
from llama_index.core.schema import TransformComponent
from pydantic import PrivateAttr
from sqlalchemy import Engine
from sqlalchemy.orm.decl_api import RegistryType
from sqlmodel.main import default_registry, Field

from autoflow.datasources import DataSource
from autoflow.datasources.mime_types import SupportedMimeTypes
from autoflow.indices.vector_search.base import VectorSearchIndex
from autoflow.storage.doc_store.base import DocumentSearchMethod
from autoflow.transformers.markdown import MarkdownNodeParser
from autoflow.schema import DataSourceType, IndexMethod, BaseComponent
from autoflow.models.chunk import get_chunk_model
from autoflow.models.entity import get_entity_model
from autoflow.models.relationship import get_relationship_model
from autoflow.knowledge_base.config import (
    ChunkingMode,
    GeneralChunkingConfig,
    ChunkSplitterConfig,
    ChunkSplitter,
    SentenceSplitterOptions,
    MarkdownNodeParserOptions,
    ChunkingConfig,
)
from autoflow.models.document import Document
from autoflow.knowledge_base.datasource import get_datasource_by_type
from autoflow.llms import default_llm_manager, LLMManager
from autoflow.llms.chat_models import ChatModel
from autoflow.llms.embeddings import EmbeddingModel
from autoflow.storage.graph_store import TiDBKnowledgeGraphStore
from autoflow.storage.doc_store import (
    TiDBDocumentStore,
    DocumentSearchResult,
)
from autoflow.storage.graph_store.base import GraphSearchAlgorithm
from autoflow.storage.schema import QueryBundle


class KnowledgeBase(BaseComponent):
    _registry: RegistryType = PrivateAttr()

    id: uuid.UUID
    name: str = Field()
    index_methods: List[IndexMethod]
    description: Optional[str] = Field(default=None)
    chunking_config: Optional[ChunkingConfig] = Field(
        default_factory=GeneralChunkingConfig
    )
    data_sources: List[DataSource] = Field(default_factory=list)

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        index_methods: Optional[List[IndexMethod]] = None,
        chunking_config: Optional[ChunkingConfig] = None,
        chat_model: Optional[ChatModel] = None,
        embedding_model: Optional[EmbeddingModel] = None,
        db_engine: Engine = None,
        llm_manager: Optional[LLMManager] = None,
        id: Optional[uuid.UUID] = None,
    ):
        super().__init__(
            id=id or uuid.uuid4(),
            name=name,
            description=description,
            index_methods=index_methods or [IndexMethod.VECTOR_SEARCH],
            chunking_config=chunking_config or GeneralChunkingConfig(),
        )
        self._db_engine = db_engine
        self._model_manager = llm_manager or default_llm_manager
        self._chat_model = chat_model

        from autoflow.utils.dspy_lm import get_dspy_lm_by_chat_model

        self._dspy_lm = get_dspy_lm_by_chat_model(self._chat_model)

        self._embedding_model = embedding_model
        self._init_stores()
        self._vector_search_index = VectorSearchIndex(
            doc_store=self._doc_store,
        )

        from autoflow.indices.knowledge_graph.base import KnowledgeGraphIndex

        self._knowledge_graph_index = KnowledgeGraphIndex(
            dspy_lm=self._dspy_lm, kg_store=self._kg_store
        )

    def _init_stores(self):
        namespace_id = f"{self.id}"
        vector_dimension = self._embedding_model.dimensions

        self._registry = RegistryType(
            metadata=default_registry.metadata,
            class_registry=default_registry._class_registry.copy(),
        )

        # Init chunk table.
        document_table_name = "documents"
        chunk_table_name = f"chunks_{namespace_id}"
        self._chunk_db_model = get_chunk_model(
            chunk_table_name,
            vector_dimension=vector_dimension,
            document_table_name=document_table_name,
            document_db_model=Document,
            registry=self._registry,
        )

        # Init entity table.
        entity_table_name = f"entities_{namespace_id}"
        self._entity_db_model = get_entity_model(
            entity_table_name,
            vector_dimension=vector_dimension,
            registry=self._registry,
        )

        # Init relationship table.
        relationship_table_name = f"relationships_{namespace_id}"
        self._relationship_db_model = get_relationship_model(
            relationship_table_name,
            vector_dimension=vector_dimension,
            entity_db_model=self._entity_db_model,
            registry=self._registry,
        )

        self._doc_store = TiDBDocumentStore[Document, self._chunk_db_model](
            db_engine=self._db_engine,
            embedding_model=self._embedding_model,
            document_db_model=Document,
            chunk_db_model=self._chunk_db_model,
        )
        self._doc_store.ensure_table_schema()

        self._kg_store = TiDBKnowledgeGraphStore(
            db_engine=self._db_engine,
            embedding_model=self._embedding_model,
            entity_db_model=self._entity_db_model,
            relationship_db_model=self._relationship_db_model,
        )
        self._kg_store.ensure_table_schema()

    def import_documents_from_datasource(
        self,
        type: DataSourceType,
        config: Dict[str, Any] = None,
        # TODO: Metadata Extractor
    ) -> DataSource:
        datasource = get_datasource_by_type(type, config)
        for doc in datasource.load_documents():
            doc.data_source_id = datasource.id
            doc.knowledge_base_id = self.id
            self.add_document(doc)
            self.build_index_for_document(doc)
        return datasource

    def import_documents_from_files(self, files: List[Path]) -> List[Document]:
        datasource = get_datasource_by_type(
            DataSourceType.FILE, {"files": [{"path": file.as_uri()} for file in files]}
        )
        documents = []
        for doc in datasource.load_documents():
            self.add_document(doc)
            self.build_index_for_document(doc)
        return documents

    def build_index_for_document(self, doc: Document):
        # Chunking
        chunks = self._chunking(doc)

        # Build Vector Search Index.
        if IndexMethod.VECTOR_SEARCH in self.index_methods:
            self._vector_search_index.build_index_for_chunks(chunks)

        # Build Knowledge Graph Index.
        if IndexMethod.KNOWLEDGE_GRAPH in self.index_methods:
            self._knowledge_graph_index.build_index_for_chunks(chunks)

    def _chunking(self, doc: Document):
        text_splitter = self._get_text_splitter(doc)
        nodes = text_splitter.get_nodes_from_documents([doc.to_llama_document()])
        return [
            self._chunk_db_model(
                hash=node.hash,
                text=node.text,
                meta={},
                document_id=doc.id,
            )
            for node in nodes
        ]

    def add_document(self, document: Document):
        return self._doc_store.add([document])

    def add_documents(self, documents: List[Document]):
        return self._doc_store.add(documents)

    def list_documents(self) -> List[Document]:
        return self._doc_store.list()

    def get_document(self, doc_id: int) -> Document:
        return self._doc_store.get(doc_id)

    def delete_document(self, doc_id: int) -> None:
        return self._doc_store.delete(doc_id)

    def _get_text_splitter(self, db_document: Document) -> TransformComponent:
        chunking_config = self.chunking_config
        if chunking_config.mode == ChunkingMode.ADVANCED:
            rules = chunking_config.rules
        else:
            rules = {
                SupportedMimeTypes.PLAIN_TXT: ChunkSplitterConfig(
                    splitter=ChunkSplitter.SENTENCE_SPLITTER,
                    splitter_options=SentenceSplitterOptions(
                        chunk_size=chunking_config.chunk_size,
                        chunk_overlap=chunking_config.chunk_overlap,
                        paragraph_separator=chunking_config.paragraph_separator,
                    ),
                ),
                SupportedMimeTypes.MARKDOWN: ChunkSplitterConfig(
                    splitter=ChunkSplitter.MARKDOWN_NODE_PARSER,
                    splitter_options=MarkdownNodeParserOptions(
                        chunk_size=chunking_config.chunk_size,
                    ),
                ),
            }

        # Chunking
        mime_type = db_document.mime_type
        if mime_type not in rules:
            raise RuntimeError(
                f"Can not chunking for the document in {db_document.mime_type} format"
            )

        rule = rules[mime_type]
        match rule.splitter:
            case ChunkSplitter.MARKDOWN_NODE_PARSER:
                options = MarkdownNodeParserOptions.model_validate(
                    rule.splitter_options
                )
                return MarkdownNodeParser(**options.model_dump())
            case ChunkSplitter.SENTENCE_SPLITTER:
                options = SentenceSplitterOptions.model_validate(rule.splitter_options)
                return SentenceSplitter(**options.model_dump())
            case _:
                raise ValueError(f"Unsupported chunking splitter type: {rule.splitter}")

    def search_documents(
        self,
        query: str,
        search_method: Optional[List[DocumentSearchMethod]] = None,
        top_k: Optional[int] = None,
        similarity_threshold: Optional[float] = None,
        similarity_nprobe: Optional[int] = None,
        similarity_top_k: Optional[int] = 5,
        **kwargs: Any,
    ) -> DocumentSearchResult:
        return self._doc_store.search(
            query=query,
            search_method=search_method or [DocumentSearchMethod.VECTOR_SEARCH],
            top_k=top_k,
            similarity_threshold=similarity_threshold,
            similarity_nprobe=similarity_nprobe,
            similarity_top_k=similarity_top_k,
            **kwargs,
        )

    def search_knowledge_graph(
        self,
        query: str,
        depth: int = 2,
        include_meta: bool = False,
        metadata_filters: Optional[dict] = None,
        search_algorithm: Optional[
            GraphSearchAlgorithm
        ] = GraphSearchAlgorithm.WEIGHTED_SEARCH,
        **kwargs,
    ):
        return self._kg_store.search(
            query=QueryBundle(query_str=query),
            depth=depth,
            include_meta=include_meta,
            metadata_filters=metadata_filters,
            search_algorithm=search_algorithm,
            **kwargs,
        )



from typing import Dict, Optional

from litellm.types.utils import LlmProviders
from llama_index.core.constants import DEFAULT_TEMPERATURE
from pydantic import BaseModel

from autoflow.llms.embeddings import EmbeddingModel
from autoflow.llms.chat_models import ChatModel
from autoflow.llms.rerankers import RerankerModel


LLMProviders = LlmProviders


class ProviderConfig(BaseModel):
    api_base: Optional[str] = None
    api_key: Optional[str] = None
    # TODO: Support Bedrock authenticate.


class ChatModelConfig(BaseModel):
    provider: str or LLMProviders
    model: str
    max_tokens: Optional[int] = None
    temperature: float = DEFAULT_TEMPERATURE
    max_retries: Optional[int] = 10


class EmbeddingModelConfig(BaseModel):
    provider: str or LLMProviders
    model: str
    dimensions: int


class RerankerModelConfig(BaseModel):
    provider: str or LLMProviders
    model: str
    top_n: Optional[int] = 5


class LLMManager:
    _registry: Dict[str, ProviderConfig] = {}

    def configure_provider(self, name: LLMProviders, config: ProviderConfig):
        self._registry[name] = config

    def get_provider(self, name: str or LLMProviders) -> Optional[ProviderConfig]:
        provider = self._registry.get(name)
        if provider is None:
            raise ValueError('Provider "{}" is not found.'.format(name))
        return provider

    def resolve_chat_model(self, config: ChatModelConfig) -> Optional[ChatModel]:
        model_provider = self.get_provider(config.provider)
        return ChatModel(
            model=f"{config.provider}/{config.model}",
            api_base=model_provider.api_base,
            api_key=model_provider.api_key,
            max_tokens=config.max_tokens,
            temperature=config.temperature,
            max_retries=config.max_retries,
        )

    def resolve_embedding_model(
        self, config: EmbeddingModelConfig
    ) -> Optional[EmbeddingModel]:
        model_provider = self.get_provider(config.provider)
        return EmbeddingModel(
            model_name=f"{config.provider}/{config.model}",
            api_base=model_provider.api_base,
            api_key=model_provider.api_key,
            dimensions=config.dimensions,
        )

    def resolve_reranker_model(
        self, config: RerankerModelConfig
    ) -> Optional[RerankerModel]:
        model_provider = self.get_provider(config.provider)
        return RerankerModel(
            model=f"{config.provider}/{config.model}",
            top_n=config.top_n,
            api_base=model_provider.api_base,
            api_key=model_provider.api_key,
        )


default_llm_manager: LLMManager = LLMManager()



from collections import defaultdict
from typing import List, Set, Tuple, Optional

from autoflow.models import EntityType
from autoflow.storage import KnowledgeGraphStore
from .base import KnowledgeGraphRetriever, RelationshipWithScore
from ..base import E, R, EntityDegree

from ...schema import QueryBundle


# The configuration for the weight coefficient
# format: ((min_weight, max_weight), coefficient)
DEFAULT_WEIGHT_COEFFICIENTS = [
    ((0, 100), 0.01),
    ((100, 1000), 0.001),
    ((1000, 10000), 0.0001),
    ((10000, float("inf")), 0.00001),
]

# The configuration for the range search
# format: ((min_distance, max_distance), search_ratio)
# The sum of search ratio should be 1 except some case we want to search as many as possible relationships.
# In this case, we set the search ratio to 1, and the other search ratio sum should be 1
DEFAULT_RANGE_SEARCH_CONFIG = [
    ((0.0, 0.25), 1),
    ((0.25, 0.35), 0.7),
    ((0.35, 0.45), 0.2),
    ((0.45, 0.55), 0.1),
]

DEFAULT_DEGREE_COEFFICIENT = 0.001


class WeightedGraphSearchRetriever(KnowledgeGraphRetriever[E, R]):
    def __init__(
        self,
        kg_store: KnowledgeGraphStore,
        with_degree: bool = False,
        alpha: float = 1,
        weight_coefficients: List[Tuple[float, float]] = None,
        search_range_config: List[Tuple[Tuple[float, float], float]] = None,
        degree_coefficient: float = DEFAULT_DEGREE_COEFFICIENT,
        fetch_synopsis_entities_num: int = 2,
        max_neighbors: int = 10,
    ):
        super().__init__(kg_store)
        self.with_degree = with_degree
        self.alpha = alpha
        self.weight_coefficients = weight_coefficients or DEFAULT_WEIGHT_COEFFICIENTS
        self.search_range_config = search_range_config or DEFAULT_RANGE_SEARCH_CONFIG
        self.degree_coefficient = degree_coefficient
        self.fetch_synopsis_entities_num = fetch_synopsis_entities_num
        self.max_neighbors = max_neighbors

    def search(
        self,
        query_embedding: List[float],
        depth: int = 2,
        metadata_filters: Optional[dict] = None,
    ) -> Tuple[List[RelationshipWithScore[R]], List[E]]:
        visited_relationships = set()
        visited_entities = set()

        new_relationships = self._weighted_search_relationships(
            query_embedding=query_embedding,
            visited_relationships=visited_relationships,
            visited_entities=visited_entities,
            metadata_filters=metadata_filters,
        )

        if len(new_relationships) == 0:
            return [], []

        for rel, score in new_relationships:
            visited_relationships.add(
                RelationshipWithScore(relationship=rel, score=score)
            )
            visited_entities.add(rel.source_entity)
            visited_entities.add(rel.target_entity)

        for _ in range(depth - 1):
            actual_number = 0
            progress = 0
            for search_config in DEFAULT_RANGE_SEARCH_CONFIG:
                search_ratio = search_config[1]
                search_distance_range = search_config[0]
                remaining_number = self.max_neighbors - actual_number
                # calculate the expected number based search progress
                # It's an accumulative search, so the expected number should be the difference between the expected number and the actual number
                expected_number = (
                    int((search_ratio + progress) * self.max_neighbors - actual_number)
                    if progress * self.max_neighbors > actual_number
                    else int(search_ratio * self.max_neighbors)
                )
                if expected_number > remaining_number:
                    expected_number = remaining_number
                if remaining_number <= 0:
                    break

                new_relationships = self._weighted_search_relationships(
                    query_embedding=query_embedding,
                    visited_relationships=visited_relationships,
                    visited_entities=visited_entities,
                    search_distance_range=search_distance_range,
                    top_k=expected_number,
                    metadata_filters=metadata_filters,
                )

                for rel, score in new_relationships:
                    visited_relationships.add(
                        RelationshipWithScore(relationship=rel, score=score)
                    )
                    visited_entities.add(rel.source_entity)
                    visited_entities.add(rel.target_entity)

                actual_number += len(new_relationships)
                # search_ratio == 1 won't count the progress
                if search_ratio != 1:
                    progress += search_ratio

        # Fetch related synopsis entities.
        synopsis_entities = self._kg_store.search_similar_entities(
            query=QueryBundle(query_embedding=query_embedding),
            entity_type=EntityType.synopsis,
            top_k=self.fetch_synopsis_entities_num,
        )
        visited_entities.update(synopsis_entities)

        return list(visited_relationships), list(visited_entities)

    def _weighted_search_relationships(
        self,
        query_embedding: List[float],
        visited_relationships: Set[RelationshipWithScore[R]],
        visited_entities: Set[E],
        search_distance_range: Tuple[float, float] = (0, 1),
        top_k: int = 10,
        metadata_filters: Optional[dict] = None,
    ) -> List[Tuple[R, float]]:
        visited_entity_ids = [e.id for e in visited_entities]
        visited_relationship_ids = [r.relationship.id for r in visited_relationships]
        relationships_with_score = self._kg_store.search_similar_relationships(
            query=QueryBundle(query_embedding=query_embedding),
            distance_range=search_distance_range,
            source_entity_ids=visited_entity_ids,
            exclude_relationship_ids=visited_relationship_ids,
            metadata_filters=metadata_filters,
        )
        return self._rerank_relationships(
            relationships_with_score=relationships_with_score,
            top_k=top_k,
        )

    def _rerank_relationships(
        self,
        relationships_with_score: List[Tuple[R, float]],
        top_k: int = 10,
    ) -> List[Tuple[R, float]]:
        """
        Rerank the relationship based on distance and weight
        """
        # TODO: the degree can br pre-calc and stored in the database in advanced.
        if self.with_degree:
            entity_ids = set()
            for r, _ in relationships_with_score:
                entity_ids.add(r.source_entity_id)
                entity_ids.add(r.target_entity_id)
            entity_degrees = self._kg_store.bulk_calc_entities_degrees(entity_ids)
        else:
            entity_degrees = defaultdict(EntityDegree)

        reranked_relationships = []
        for r, similarity_score in relationships_with_score:
            embedding_distance = 1 - similarity_score
            source_in_degree = entity_degrees[r.source_entity_id].in_degree
            target_out_degree = entity_degrees[r.target_entity_id].out_degree
            final_score = self._calc_relationship_weighted_score(
                embedding_distance,
                r.weight,
                source_in_degree,
                target_out_degree,
            )
            reranked_relationships.append((r, final_score))

        # Rerank relationships based on the calculated score.
        reranked_relationships.sort(key=lambda x: x[1], reverse=True)
        return reranked_relationships[:top_k]

    def _calc_relationship_weighted_score(
        self, embedding_distance: float, weight: int, in_degree: int, out_degree: int
    ) -> float:
        weighted_score = self._calc_weight_score(weight)
        degree_score = 0
        if self.with_degree:
            degree_score = self._calc_degree_score(in_degree, out_degree)
        return self.alpha * (1 / embedding_distance) + weighted_score + degree_score

    def _calc_weight_score(self, weight: float) -> float:
        weight_score = 0.0
        remaining_weight = weight

        for weight_range, coefficient in self.weight_coefficients:
            if remaining_weight <= 0:
                break
            lower_bound, upper_bound = weight_range
            applicable_weight = min(upper_bound - lower_bound, remaining_weight)
            weight_score += applicable_weight * coefficient
            remaining_weight -= applicable_weight

        return weight_score

    def _calc_degree_score(self, in_degree: int, out_degree: int) -> float:
        return (in_degree - out_degree) * self.degree_coefficient



import logging
from pydantic import BaseModel
from typing import Generator, List

from autoflow.datasources import DataSource
from autoflow.datasources.web_base import load_web_documents
from autoflow.models import DBDocument

logger = logging.getLogger(__name__)


class WebSinglePageDataSourceConfig(BaseModel):
    urls: List[str]


class WebSinglePageDataSource(DataSource[WebSinglePageDataSourceConfig]):
    def validate_config(self, config: dict):
        return WebSinglePageDataSourceConfig.model_validate(config)

    def load_documents(self) -> Generator[DBDocument, None, None]:
        for doc in load_web_documents(self.config.urls):
            doc.data_source_id = self.id
            yield doc



DEFAULT_INTENT_GRAPH_KNOWLEDGE = """\
Given a list of prerequisite questions and their relevant knowledge for the user's main question, when conflicts in meaning arise, prioritize the relationship with the higher weight and the more recent version.

Knowledge sub-queries:

{% for sub_query, data in sub_queries.items() %}

Sub-query: {{ sub_query }}

  - Entities:

{% for entity in data['entities'] %}

    - Name: {{ entity.name }}
    - Description: {{ entity.description }}

{% endfor %}

  - Relationships:

{% for relationship in data['relationships'] %}

    - Description: {{ relationship.rag_description }}
    - Weight: {{ relationship.weight }}

{% endfor %}

{% endfor %}
"""

DEFAULT_NORMAL_GRAPH_KNOWLEDGE = """\
Given a list of relationships of a knowledge graph as follows. When there is a conflict in meaning between knowledge relationships, the relationship with the higher `weight` and newer `last_modified_at` value takes precedence.

---------------------
Entities:

{% for entity in entities %}

- Name: {{ entity.name }}
- Description: {{ entity.description }}

{% endfor %}

---------------------

Knowledge relationships:

{% for relationship in relationships %}

- Description: {{ relationship.rag_description }}
- Weight: {{ relationship.weight }}
- Last Modified At: {{ relationship.last_modified_at }}
- Meta: {{ relationship.meta | tojson(indent=2) }}

{% endfor %}
"""

DEFAULT_CLARIFYING_QUESTION_PROMPT = """\
---------------------
The prerequisite questions and their relevant knowledge for the user's main question.
---------------------

{{graph_knowledges}}

---------------------

Task:
Given the conversation between the user and ASSISTANT, along with the follow-up message from the user, and the provided prerequisite questions and relevant knowledge, determine if the user's question is clear and specific enough for a confident response. If the question lacks necessary details or context, identify the specific ambiguities and generate a clarifying question to address them.

Instructions:
1. Assess Information Sufficiency:
   - Evaluate if the user’s question provides enough detail to generate a precise answer based on the prerequisite questions, relevant knowledge, and conversation history.
   - If the user's question is too vague or lacks key information, identify what additional information would be necessary for clarity.

2. Generate a Clarifying Question:
   - If the question is clear and answerable, return exact "False" as the response.
   - If clarification is needed, return a specific question to ask the user, directly addressing the information gap. Avoid general questions; focus on the specific details required for an accurate answer.

3. Use the same language to ask the clarifying question as the user's original question.

Example 1:

user: "Does TiDB support foreign keys?"
Relevant Knowledge: TiDB supports foreign keys starting from version 6.6.0.

Response:

Which version of TiDB are you using?

Example 2:

user: "Does TiDB support nested transaction?"
Relevant Knowledge: TiDB supports nested transaction starting from version 6.2.0.

Response:

Which version of TiDB are you using?

Example 3:

user: "Does TiDB support foreign keys? I'm using TiDB 6.5.0."
Relevant Knowledge: TiDB supports foreign keys starting from version 6.6.0.

Response:

False

Your Turn:

Chat history:

{{chat_history}}

---------------------

Follow-up question:

{{question}}

Response:
"""

DEFAULT_CONDENSE_QUESTION_PROMPT = """\
Current Date: {{current_date}}
---------------------
The prerequisite questions and their relevant knowledge for the user's main question.
---------------------

{{graph_knowledges}}

---------------------

Task:
Given the conversation between the Human and Assistant, along with the follow-up message from the Human, and the provided prerequisite questions and relevant knowledge, refine the Human’s follow-up message into a standalone, detailed question.

Instructions:
1. Focus on the latest query from the Human, ensuring it is given the most weight.
2. Incorporate Key Information:
  - Use the prerequisite questions and their relevant knowledge to add specific details to the follow-up question.
  - Replace ambiguous terms or references in the follow-up question with precise information from the provided knowledge. Example: Replace “latest version” with the actual version number mentioned in the knowledge.
3. Utilize Conversation Context:
  - Incorporate relevant context and background information from the conversation history to enhance the question's specificity.
4. Optimize for Retrieval:
  - Ensure the refined question emphasizes specific and relevant terms to maximize the effectiveness of a vector search for retrieving precise and comprehensive information.
5. Grounded and Factual:
  - Make sure the refined question is grounded in and directly based on the user's follow-up question and the provided knowledge.
  - Do not introduce information that is not supported by the knowledge or conversation history.
6. Give the language hint for the answer:
  - Add a hint after the question like "(Answer language: English)", or "(Answer language: Chinese)", etc.
  - This language hint should be exactly same with the language of the original question.
  - If the original question has part of other language aside from English, please use the language of another language rather than English. Example: "tidb tableread慢会是哪些原因", it should be Chinese.

Example:

Chat History:

Human: "I'm interested in the performance improvements in the latest version of TiDB."
Assistant: "TiDB version 8.1 was released recently with significant performance enhancements over version 6.5."

Follow-up Question:

"Can you tell me more about these improvements?"

Prerequisite Questions and Relevant Knowledge:

- Prerequisite Question: What is the latest version of TiDB?
- Relevant Knowledge: The latest version of TiDB is 8.1.

...

Refined Standalone Question:

"Can you provide detailed information about the performance improvements introduced in TiDB version 8.1 compared to version 6.5? (Answer language: English)"

Your Turn:

Chat history:

{{chat_history}}

---------------------

Followup question:

{{question}}

---------------------

Refined standalone question:
"""

DEFAULT_TEXT_QA_PROMPT = """\
Current Date: {{current_date}}
---------------------
Knowledge graph information is below
---------------------

{{graph_knowledges}}

---------------------
Context information is below.
---------------------

<<context_str>>

---------------------

Answer Format:

Use markdown footnote syntax (for example: [^1]) to indicate sources you used.
Each footnote must correspond to a unique source. Do not use the same source for multiple footnotes.

### Examples of Correct Footnote Usage (no the unique sources and diverse sources):
[^1]: [TiDB Overview | PingCAP Docs](https://docs.pingcap.com/tidb/stable/overview)
[^2]: [TiDB Architecture | PingCAP Docs](https://docs.pingcap.com/tidb/stable/architecture)

### Examples of Incorrect Footnote Usage (Avoid duplicating the same source for multiple footnotes):
[^1]: [TiDB Introduction | PingCAP Docs](https://docs.pingcap.com/tidb/v5.4/overview)
[^2]: [TiDB Introduction | PingCAP Docs](https://docs.pingcap.com/tidb/v5.4/overview)
[^3]: [TiDB Introduction | PingCAP Docs](https://docs.pingcap.com/tidb/dev/overview)
[^4]: [TiDB Introduction | PingCAP Docs](https://docs.pingcap.com/tidb/stable/overview)

---------------------

Answer Language:

Follow the language of the language hint after the Refined Question.
If the language hint is not provided, use the language that the original questions used.

---------------------

As a customer support assistant, please do not fabricate any knowledge. If you cannot get knowledge from the context, please just directly state "you do not know", rather than constructing nonexistent and potentially fake information!!!

First, analyze the provided context information without assuming prior knowledge. Identify all relevant aspects of knowledge contained within. Then, from various perspectives and angles, answer questions as thoroughly and comprehensively as possible to better address and resolve the user's issue.

The Original questions is:

{{original_question}}

The Refined Question used to search:
<<query_str>>

Answer:
"""

DEFAULT_FURTHER_QUESTIONS_PROMPT = """\
The chat message content is:

{{chat_message_content}}

---------------------
Task:
Based on the provided chat message, generate 3–5 follow-up questions that are relevant to the content. Each question should explore the topic in greater detail, seek clarification, or introduce new angles for discussion.

Instructions:
1. Build upon the key information, themes, or insights within the provided chat message.
2. Aim for variety in question type (clarifying, probing, or exploratory) to encourage a deeper conversation.
3. Ensure each question logically follows from the context of the provided chat message.
4. Keep questions concise yet insightful to maximize engagement.
5. Use the same language with the chat message content.
6. Each question should end with a question mark.
7. Each question should be in a new line, DO NOT add any indexes or blank lines, just output the questions.

Now, generate 3–5 follow-up questions below:
"""



from abc import ABC, abstractmethod
from typing import List

from autoflow.storage.tidb.base import TiDBModel


class BaseRerankFunction(ABC):
    @abstractmethod
    def rerank(
        self, items: List[TiDBModel], query_str: str, top_n: int = 2
    ) -> List[TiDBModel]:
        pass



from .provider import (
    LLMProviders,
    ProviderConfig,
    LLMManager,
    default_llm_manager,
    ChatModelConfig,
    EmbeddingModelConfig,
    RerankerModelConfig,
)

__all__ = [
    "LLMProviders",
    "ProviderConfig",
    "LLMManager",
    "default_llm_manager",
    "ChatModelConfig",
    "EmbeddingModelConfig",
    "RerankerModelConfig",
]



from abc import ABC, abstractmethod
from typing import Optional

from sqlalchemy import Column
from sqlmodel import Field
from tidb_vector.sqlalchemy import VectorType


class BaseEmbeddingFunction(ABC):
    def VectorField(self, source_field: Optional[str] = None, **kwargs):
        dimensions = self._get_dimensions()
        return Field(
            sa_column=Column(VectorType(self._get_dimensions())),
            schema_extra={
                "embed_fn": self,
                "dimensions": dimensions,
                "source_field": source_field,
            },
            **kwargs,
        )

    @abstractmethod
    def _get_dimensions(self) -> int:
        pass

    @abstractmethod
    def get_query_embedding(self, query: str) -> list[float]:
        pass

    @abstractmethod
    def get_source_embedding(self, source: str) -> list[float]:
        pass

    @abstractmethod
    def get_source_embedding_batch(self, source: str) -> list[float]:
        pass



from pydantic import BaseModel, Field
from typing import Mapping, Any, List


class AIEntity(BaseModel):
    """List of entities extracted from the text to form the knowledge graph"""

    name: str = Field(
        description="Name of the entity, it should be a clear and concise term"
    )
    description: str = Field(
        description=(
            "Description of the entity, it should be a complete and comprehensive sentence, not few words. "
            "Sample description of entity 'TiDB in-place upgrade': "
            "'Upgrade TiDB component binary files to achieve upgrade, generally use rolling upgrade method'"
        )
    )
    metadata: Mapping[str, Any] = Field(
        description=(
            "The covariates (which is a comprehensive json TREE, the first field is always: 'topic') to claim the entity. "
        )
    )


class AIEntityWithID(AIEntity):
    """Entity extracted from the text to form the knowledge graph with an ID."""

    id: int = Field(description="Unique identifier for the entity.")


class SynopsisInfo(BaseModel):
    """A synopsis corresponds to a group of entities that share the same topic and can contribute to synopsis topic."""

    topic: str = Field(
        description="The shared topic of the synopsis, and each entity in the group can contribute factual data from its own perspective."
    )
    entities: List[int] = Field(
        description="A group of entity(only IDs) that can contribute to the synopsis base on the analysis of entity descriptions and metadata."
    )


class AISynopsisEntity(AIEntity):
    """Unified synopsis entity with comprehensive description and metadata based on the entities group."""

    group_info: SynopsisInfo = Field(
        description="Group of entities to be unified into a single synopsis entity."
    )


class ExistingSynopsisEntity(AISynopsisEntity):
    """Unified synopsis entity with comprehensive description and metadata based on the entities group."""

    id: int = Field(description="Unique identifier for the entity.")


class AIRelationship(BaseModel):
    """List of relationships extracted from the text to form the knowledge graph"""

    source_entity: str = Field(
        description="Source entity name of the relationship, it should an existing entity in the Entity list"
    )
    target_entity: str = Field(
        description="Target entity name of the relationship, it should an existing entity in the Entity list"
    )
    relationship_desc: str = Field(
        description=(
            "Description of the relationship, it should be a complete and comprehensive sentence, not few words. "
            "Sample relationship description: 'TiDB will release a new LTS version every 6 months.'"
        )
    )


class AIRelationshipWithEntityDesc(AIRelationship):
    source_entity_description: str = Field()
    target_entity_description: str = Field()
    meta: Mapping[str, Any] = Field()


class AIRelationshipReasoning(AIRelationship):
    """Relationship between two entities extracted from the query"""

    reasoning: str = Field(
        description=(
            "Category reasoning for the relationship, e.g., 'the main conerns of the user', 'the problem the user is facing', 'the user case scenario', etc."
        )
    )


class AIKnowledgeGraph(BaseModel):
    """Graph representation of the knowledge for text."""

    entities: List[AIEntity] = Field(
        description="List of entities in the knowledge graph"
    )
    relationships: List[AIRelationship] = Field(
        description="List of relationships in the knowledge graph"
    )


class ProcessedAIKnowledgeGraph(BaseModel):
    entities: List[AIEntity]
    relationships: List[AIRelationshipWithEntityDesc]


class EntityCovariateInput(BaseModel):
    """List of entities extracted from the text to form the knowledge graph"""

    name: str = Field(description="Name of the entity")
    description: str = Field(description="Description of the entity")


class EntityCovariateOutput(BaseModel):
    """List of entities extracted from the text to form the knowledge graph"""

    name: str = Field(description="Name of the entity")
    description: str = Field(description="Description of the entity")
    covariates: Mapping[str, Any] = Field(
        description=(
            "The attributes (which is a comprehensive json TREE, the first field is always: 'topic') to claim the entity. "
        )
    )


class DecomposedFactors(BaseModel):
    """Decomposed factors extracted from the query to form the knowledge graph"""

    relationships: List[AIRelationshipReasoning] = Field(
        description="List of relationships to represent critical concepts and their relationships extracted from the query."
    )



from typing import (
    Any,
    Dict,
    List,
    Set,
    Tuple,
    Type,
    cast,
)

from sqlalchemy import (
    inspect,
)
from sqlalchemy.orm import (
    Mapped,
    registry,
    relationship,
)
from sqlalchemy.orm.decl_api import DeclarativeMeta
from sqlmodel.main import (
    BaseConfig,
    ModelField,
    ModelMetaclass,
    SQLModelConfig,
    Undefined,
    get_annotations,
    get_config_value,
    get_model_fields,
    get_relationship_to,
    is_table_model_class,
    set_config_value,
    __dataclass_transform__,
    Field,
    FieldInfo,
    RelationshipInfo,
    get_column_from_field,
)
from typing_extensions import get_origin


# Using CustomSQLModel and CustomSQLModelMetaclass, the default CustomSQLModelMetaclass will mark this sqlmodel
# as __abstract__=True if you custom registry via kwargs, which will avoid it registering class into SQLAlchemy,
# and then you will receive error while using some sqlalchemy feature.
@__dataclass_transform__(kw_only_default=True, field_descriptors=(Field, FieldInfo))
class PatchSQLModelMetaclass(ModelMetaclass, DeclarativeMeta):
    __sqlmodel_relationships__: Dict[str, RelationshipInfo]
    model_config: SQLModelConfig
    model_fields: Dict[str, FieldInfo]
    __config__: Type[SQLModelConfig]
    __fields__: Dict[str, ModelField]  # type: ignore[assignment]

    # Replicate SQLAlchemy
    def __setattr__(cls, name: str, value: Any) -> None:
        if is_table_model_class(cls):
            DeclarativeMeta.__setattr__(cls, name, value)
        else:
            super().__setattr__(name, value)

    def __delattr__(cls, name: str) -> None:
        if is_table_model_class(cls):
            DeclarativeMeta.__delattr__(cls, name)
        else:
            super().__delattr__(name)

    # From Pydantic
    def __new__(
        cls,
        name: str,
        bases: Tuple[Type[Any], ...],
        class_dict: Dict[str, Any],
        **kwargs: Any,
    ) -> Any:
        relationships: Dict[str, RelationshipInfo] = {}
        dict_for_pydantic = {}
        original_annotations = get_annotations(class_dict)
        pydantic_annotations = {}
        relationship_annotations = {}
        for k, v in class_dict.items():
            if isinstance(v, RelationshipInfo):
                relationships[k] = v
            else:
                dict_for_pydantic[k] = v
        for k, v in original_annotations.items():
            if k in relationships:
                relationship_annotations[k] = v
            else:
                pydantic_annotations[k] = v
        dict_used = {
            **dict_for_pydantic,
            "__weakref__": None,
            "__sqlmodel_relationships__": relationships,
            "__annotations__": pydantic_annotations,
        }
        # Duplicate logic from Pydantic to filter config kwargs because if they are
        # passed directly including the registry Pydantic will pass them over to the
        # superclass causing an error
        allowed_config_kwargs: Set[str] = {
            key
            for key in dir(BaseConfig)
            if not (
                key.startswith("__") and key.endswith("__")
            )  # skip dunder methods and attributes
        }
        config_kwargs = {
            key: kwargs[key] for key in kwargs.keys() & allowed_config_kwargs
        }
        new_cls = super().__new__(cls, name, bases, dict_used, **config_kwargs)
        new_cls.__annotations__ = {
            **relationship_annotations,
            **pydantic_annotations,
            **new_cls.__annotations__,
        }

        def get_config(name: str) -> Any:
            config_class_value = get_config_value(
                model=new_cls, parameter=name, default=Undefined
            )
            if config_class_value is not Undefined:
                return config_class_value
            kwarg_value = kwargs.get(name, Undefined)
            if kwarg_value is not Undefined:
                return kwarg_value
            return Undefined

        config_table = get_config("table")
        if config_table is True:
            # If it was passed by kwargs, ensure it's also set in config
            set_config_value(model=new_cls, parameter="table", value=config_table)
            for k, v in get_model_fields(new_cls).items():
                col = get_column_from_field(v)
                setattr(new_cls, k, col)
            # Set a config flag to tell FastAPI that this should be read with a field
            # in orm_mode instead of preemptively converting it to a dict.
            # This could be done by reading new_cls.model_config['table'] in FastAPI, but
            # that's very specific about SQLModel, so let's have another config that
            # other future tools based on Pydantic can use.
            set_config_value(
                model=new_cls, parameter="read_from_attributes", value=True
            )
            # For compatibility with older versions
            # TODO: remove this in the future
            set_config_value(model=new_cls, parameter="read_with_orm_mode", value=True)

        config_registry = get_config("registry")
        if config_registry is not Undefined:
            config_registry = cast(registry, config_registry)
            # If it was passed by kwargs, ensure it's also set in config
            set_config_value(model=new_cls, parameter="registry", value=config_table)
            setattr(new_cls, "_sa_registry", config_registry)  # noqa: B010
            setattr(new_cls, "metadata", config_registry.metadata)  # noqa: B010
            setattr(new_cls, "__abstract__", False)  # noqa: B010                           <- Using True instead of False

        return new_cls

    # Override SQLAlchemy, allow both SQLAlchemy and plain Pydantic models
    def __init__(
        cls, classname: str, bases: Tuple[type, ...], dict_: Dict[str, Any], **kw: Any
    ) -> None:
        # Only one of the base classes (or the current one) should be a table model
        # this allows FastAPI cloning a SQLModel for the response_model without
        # trying to create a new SQLAlchemy, for a new table, with the same name, that
        # triggers an error
        base_is_table = any(is_table_model_class(base) for base in bases)
        if is_table_model_class(cls) and not base_is_table:
            for rel_name, rel_info in cls.__sqlmodel_relationships__.items():
                if rel_info.sa_relationship:
                    # There's a SQLAlchemy relationship declared, that takes precedence
                    # over anything else, use that and continue with the next attribute
                    setattr(cls, rel_name, rel_info.sa_relationship)  # Fix #315
                    continue
                raw_ann = cls.__annotations__[rel_name]
                origin = get_origin(raw_ann)
                if origin is Mapped:
                    ann = raw_ann.__args__[0]
                else:
                    ann = raw_ann
                    # Plain forward references, for models not yet defined, are not
                    # handled well by SQLAlchemy without Mapped, so, wrap the
                    # annotations in Mapped here
                    cls.__annotations__[rel_name] = Mapped[ann]  # type: ignore[valid-type]
                relationship_to = get_relationship_to(
                    name=rel_name, rel_info=rel_info, annotation=ann
                )
                rel_kwargs: Dict[str, Any] = {}
                if rel_info.back_populates:
                    rel_kwargs["back_populates"] = rel_info.back_populates
                if rel_info.link_model:
                    ins = inspect(rel_info.link_model)
                    local_table = getattr(ins, "local_table")  # noqa: B009
                    if local_table is None:
                        raise RuntimeError(
                            "Couldn't find the secondary table for "
                            f"model {rel_info.link_model}"
                        )
                    rel_kwargs["secondary"] = local_table
                rel_args: List[Any] = []
                if rel_info.sa_relationship_args:
                    rel_args.extend(rel_info.sa_relationship_args)
                if rel_info.sa_relationship_kwargs:
                    rel_kwargs.update(rel_info.sa_relationship_kwargs)
                rel_value = relationship(relationship_to, *rel_args, **rel_kwargs)
                setattr(cls, rel_name, rel_value)  # Fix #315
            # SQLAlchemy no longer uses dict_
            # Ref: https://github.com/sqlalchemy/sqlalchemy/commit/428ea01f00a9cc7f85e435018565eb6da7af1b77
            # Tag: 1.4.36
            DeclarativeMeta.__init__(cls, classname, bases, dict_, **kw)
        else:
            ModelMetaclass.__init__(cls, classname, bases, dict_, **kw)



from uuid import UUID
from sqlmodel import Field, SQLModel
from sqlalchemy.types import TypeDecorator, Integer

from autoflow.utils.uuid6 import uuid7


class UUIDBaseModel(SQLModel):
    id: UUID = Field(
        default_factory=uuid7,
        primary_key=True,
        index=True,
        nullable=False,
        sa_column_kwargs={"sort_order": -1},
    )


class IntEnumType(TypeDecorator):
    """
    IntEnumType is a custom TypeDecorator that handles conversion between
    integer values in the database and Enum types in Python.

    This replaces the previous SmallInteger implementation to resolve Pydantic
    serialization warnings. When using SmallInteger, SQLAlchemy would return raw
    integers from the database (e.g., 0 or 1), causing Pydantic validation warnings
    since it expects proper Enum types.
    """

    impl = Integer

    def __init__(self, enum_class, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enum_class = enum_class

    def process_bind_param(self, value, dialect):
        # enum -> int
        if isinstance(value, self.enum_class):
            return value.value
        elif value is None:
            return None
        raise ValueError(f"Invalid value for {self.enum_class}: {value}")

    def process_result_value(self, value, dialect):
        # int -> enum
        if value is not None:
            return self.enum_class(value)
        return None



from typing import List, Optional

from litellm.rerank_api.main import rerank
from llama_index.core.bridge.pydantic import Field
from llama_index.core.callbacks import CBEventType, EventPayload
from llama_index.core.instrumentation import get_dispatcher
from llama_index.core.instrumentation.events.rerank import (
    ReRankEndEvent,
    ReRankStartEvent,
)
from llama_index.core.postprocessor.types import BaseNodePostprocessor
from llama_index.core.schema import NodeWithScore, QueryBundle, MetadataMode

dispatcher = get_dispatcher(__name__)


class LiteLLMReranker(BaseNodePostprocessor):
    model: str = Field(description="Reranker model name.")
    top_n: int = Field(description="Top N nodes to return.")
    api_base: Optional[str] = Field(description="Reranker API base url.", default=None)
    api_key: Optional[str] = Field(description="Reranker API key.")

    def __init__(
        self,
        top_n: int = 2,
        model: str = "rerank-english-v2.0",
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
    ):
        super().__init__(top_n=top_n, model=model, api_base=api_base, api_key=api_key)

    @classmethod
    def class_name(cls) -> str:
        return "LiteLLMRerank"

    def _postprocess_nodes(
        self,
        nodes: List[NodeWithScore],
        query_bundle: Optional[QueryBundle] = None,
    ) -> List[NodeWithScore]:
        dispatcher.event(
            ReRankStartEvent(
                query=query_bundle, nodes=nodes, top_n=self.top_n, model_name=self.model
            )
        )

        if query_bundle is None:
            raise ValueError("Missing query bundle in extra info.")
        if len(nodes) == 0:
            return []

        with self.callback_manager.event(
            CBEventType.RERANKING,
            payload={
                EventPayload.NODES: nodes,
                EventPayload.MODEL_NAME: self.model,
                EventPayload.QUERY_STR: query_bundle.query_str,
                EventPayload.TOP_K: self.top_n,
            },
        ) as event:
            texts = [
                node.node.get_content(metadata_mode=MetadataMode.EMBED)
                for node in nodes
            ]
            results = rerank(
                model=self.model,
                query=query_bundle.query_str,
                documents=texts,
                top_n=self.top_n,
                api_base=self.api_base,
                api_key=self.api_key,
            )

            new_nodes = []
            for result in results.results:
                new_node_with_score = NodeWithScore(
                    node=nodes[result["index"]].node, score=result["relevance_score"]
                )
                new_nodes.append(new_node_with_score)
            event.on_end(payload={EventPayload.NODES: new_nodes})

        dispatcher.event(ReRankEndEvent(nodes=new_nodes))
        return new_nodes



import uuid
from typing import Optional, List, TYPE_CHECKING

import sqlalchemy
from sqlalchemy.engine import Engine
from sqlmodel import SQLModel
from autoflow.llms.chat_models import ChatModel
from autoflow.llms.embeddings import EmbeddingModel
from autoflow.schema import IndexMethod
from autoflow.storage.tidb.utils import build_tidb_dsn

if TYPE_CHECKING:
    from autoflow.llms import LLMManager
    from autoflow.knowledge_base import KnowledgeBase


class Autoflow:
    _db_engine = None
    _model_manager = None

    def __init__(self, db_engine: Engine, model_manager: Optional[LLMManager] = None):
        from autoflow.llms import default_llm_manager

        self._db_engine = db_engine
        self._model_manager = model_manager or default_llm_manager
        self._init_table_schema()

    @classmethod
    def from_config(
        cls,
        db_host: Optional[str] = None,
        db_port: Optional[int] = None,
        db_username: Optional[str] = None,
        db_password: Optional[str] = None,
        db_name: str = "autoflow",
        db_enable_ssl: Optional[bool] = None,
    ):
        tidb_dsn = str(
            build_tidb_dsn(
                host=db_host,
                port=db_port,
                username=db_username,
                password=db_password,
                database=db_name,
                enable_ssl=db_enable_ssl,
            )
        )
        db_engine = sqlalchemy.create_engine(str(tidb_dsn))
        return cls(
            db_engine=db_engine,
        )

    def _init_table_schema(self):
        SQLModel.metadata.create_all(self._db_engine)

    @property
    def db_engine(self) -> Engine:
        return self._db_engine

    @property
    def llm_manager(self) -> LLMManager:
        return self._model_manager

    def create_knowledge_base(
        self,
        name: str,
        chat_model: ChatModel,
        embedding_model: EmbeddingModel,
        description: Optional[str] = None,
        index_methods: Optional[List[IndexMethod]] = None,
        id: Optional[uuid.UUID] = None,
    ) -> KnowledgeBase:
        return KnowledgeBase(
            name=name,
            description=description,
            index_methods=index_methods,
            chat_model=chat_model,
            embedding_model=embedding_model,
            id=id,
            db_engine=self._db_engine,
        )
